import traceback
from queue import Full
from .task import Producer
from actfw_core.v4l2.video import Video, VideoPort, V4L2_PIX_FMT
import enum


class Frame(object):

    """Captured Frame"""

    def __init__(self, value):
        self.value = value
        self.updatable = True

    def getvalue(self):
        """
        Get frame data.

        Returns:
            bytes: captured image data

        """
        self.updatable = False
        return self.value

    def _update(self, value):
        if self.updatable:
            self.value = value
            return True
        return False


class V4LCameraCapture(Producer):

    FormatSelector = enum.Enum('FormatSelector', 'DEFAULT PROPER MAXIMUM')

    """Captured Frame Producer for Video4Linux"""

    def __init__(self, device='/dev/video0', size=(640, 480), framerate=30,
                 expected_format=V4L2_PIX_FMT.RGB24,
                 fallback_formats=[V4L2_PIX_FMT.YUYV, V4L2_PIX_FMT.MJPEG],
                 format_selector=FormatSelector.DEFAULT):
        """

        Args:
            device (str): v4l device path
            size (int, int): expected capture resolution
            framerate (int): expected capture framerate
            expected_format (:class:`~actfw_core.v4l2.video.V4L2_PIX_FMT`): expected capture format
            fallback_formats (list of :class:`~actfw_core.v4l2.video.V4L2_PIX_FMT`): fallback capture format
            format_selector (:class:`~actfw_core.capture.V4LCameraCapture.FormatSelector): how to select a format from listed formats supported by a camera.
                DEFAULT selects the first format that meets the conditions.
                PROPER selects the smallest format that meets the conditions.
                MAXIMUM selects the largest resolution format as a camera can.
                **MAXIMUM ignores framerate parameters (uses appropriate framerate for the selected resolution).**
                If a camera lists [1280x720, 1920x1080, 640x480, 800x600] and an expected capture resolution is (512, 512),
                DEFAULT selects 1280x720, PROPER selects 800x600 and MAXIMUM selects 1920x1080.

        Notes:
            If a camera doesn't support the expected_format,
            try to capture one of the fallback_formats and convert it to expected_format.

        """
        super(V4LCameraCapture, self).__init__()
        self.video = Video(device)
        self.frames = []

        width, height = size

        if self.video.query_capability() == VideoPort.CSI:

            # workaround for bcm2835-v4l2 format pixsize & bytesperline bug
            width = (width + 31) // 32 * 32
            height = (height + 15) // 16 * 16

            # workaround for bcm2835-v4l2 IMX219 32x32 -> 800x800 capture timeout bug
            candidates = self.video.lookup_config(64, 64, 5, V4L2_PIX_FMT.RGB24, V4L2_PIX_FMT.RGB24)
            self.video.set_format(candidates[0], 64, 64, V4L2_PIX_FMT.RGB24)

        if format_selector in [V4LCameraCapture.FormatSelector.PROPER, V4LCameraCapture.FormatSelector.MAXIMUM]:
            def cmp(config):
                return (
                    config.width * config.height,
                    config.height,
                    config.width,
                    config.interval.denominator / config.interval.numerator
                )
        else:
            def cmp(config):
                return 1
        config = None
        fmts = [expected_format] + fallback_formats
        for fmt in fmts:
            expected_framerate = 1 if format_selector == V4LCameraCapture.FormatSelector.MAXIMUM else framerate
            candidates = self.video.lookup_config(width, height, expected_framerate, fmt, expected_format)
            candidates = sorted(candidates, key=cmp)
            if len(candidates) > 0:
                config = candidates[-1 if format_selector == V4LCameraCapture.FormatSelector.MAXIMUM else 0]
                break
        if config is None:
            raise RuntimeError("expected capture format is unsupported")
        if format_selector == V4LCameraCapture.FormatSelector.MAXIMUM:
            fmt = self.video.set_format(config, expected_format=expected_format)
        else:
            fmt = self.video.set_format(config, width, height, expected_format=expected_format)
        self.capture_width, self.capture_height, self.capture_format = fmt
        self.video.set_framerate(config)
        # video.set_rotation(90)
        buffers = self.video.request_buffers(4)
        for buf in buffers:
            self.video.queue_buffer(buf)

    def capture_size(self):
        """
        Get configured capture resolution.
        A configured resolution may be more larger than expected one.

        Returns:
            (int, int): configured capture resolution (width, height)
        """
        return (self.capture_width, self.capture_height)

    def configure(self, configurator):
        """
        Run user defined video configurator.

        Args:
            configurator : unary function (`actfw_core.v4l2.video.Video` -> a)

        Returns:
            object: return type of configurator
        """
        return configurator(self.video)

    def run(self):
        """Run producer activity"""
        with self.video.start_streaming() as stream:
            while self._is_running():
                try:
                    value = stream.capture(timeout=5)
                    updated = 0
                    for frame in reversed(self.frames):
                        if frame._update(value):
                            updated += 1
                        else:
                            break
                    self.frames = self.frames[len(self.frames) - updated:]
                    frame = Frame(value)
                    if self._outlet(frame):
                        self.frames.append(frame)
                except:
                    raise
        self.video.close()

    def _outlet(self, o):
        length = len(self.out_queues)
        while self._is_running():
            try:
                self.out_queues[self.out_queue_id % length].put(o, block=False)
                self.out_queue_id += 1
                return True
            except Full:
                return False
            except:
                traceback.print_exc()
        return False



import enum


class V4L2_CTRL_CLASS(enum.IntEnum):
    USER = 0x00980000
    MPEG = 0x00990000
    CAMERA = 0x009a0000
    FM_TX = 0x009b0000
    FLASH = 0x009c0000
    JPEG = 0x009d0000
    IMAGE_SOURCE = 0x009e0000
    IMAGE_PROC = 0x009f0000
    DV = 0x00a00000
    FM_RX = 0x00a10000
    RF_TUNER = 0x00a20000
    DETECT = 0x00a30000


class V4L2_CID(enum.IntEnum):
    BASE = (V4L2_CTRL_CLASS.USER | 0x900)
    USER_BASE = BASE
    USER_CLASS = (V4L2_CTRL_CLASS.USER | 1)
    BRIGHTNESS = (BASE+0)
    CONTRAST = (BASE+1)
    SATURATION = (BASE+2)
    HUE = (BASE+3)

    AUDIO_VOLUME = (BASE+5)
    AUDIO_BALANCE = (BASE+6)
    AUDIO_BASS = (BASE+7)
    AUDIO_TREBLE = (BASE+8)
    AUDIO_MUTE = (BASE+9)
    AUDIO_LOUDNESS = (BASE+10)

    BLACK_LEVEL = (BASE+11)
    AUTO_WHITE_BALANCE = (BASE+12)
    DO_WHITE_BALANCE = (BASE+13)
    RED_BALANCE = (BASE+14)
    BLUE_BALANCE = (BASE+15)
    GAMMA = (BASE+16)
    WHITENESS = (GAMMA)
    EXPOSURE = (BASE+17)
    AUTOGAIN = (BASE+18)
    GAIN = (BASE+19)
    HFLIP = (BASE+20)
    VFLIP = (BASE+21)
    POWER_LINE_FREQUENCY = (BASE+24)
    HUE_AUTO = (BASE+25)
    WHITE_BALANCE_TEMPERATURE = (BASE+26)
    SHARPNESS = (BASE+27)
    BACKLIGHT_COMPENSATION = (BASE+28)
    CHROMA_AGC = (BASE+29)
    COLOR_KILLER = (BASE+30)
    COLORFX = (BASE+31)
    AUTOBRIGHTNESS = (BASE+32)
    BAND_STOP_FILTER = (BASE+33)
    ROTATE = (BASE+34)
    BG_COLOR = (BASE+35)
    CHROMA_GAIN = (BASE+36)
    ILLUMINATORS_1 = (BASE+37)
    ILLUMINATORS_2 = (BASE+38)
    MIN_BUFFERS_FOR_CAPTURE = (BASE+39)
    MIN_BUFFERS_FOR_OUTPUT = (BASE+40)
    ALPHA_COMPONENT = (BASE+41)
    COLORFX_CBCR = (BASE+42)
    LASTP1 = (BASE+43)

    USER_MEYE_BASE = (USER_BASE + 0x1000)
    USER_BTTV_BASE = (USER_BASE + 0x1010)
    USER_S2255_BASE = (USER_BASE + 0x1030)
    USER_SI476X_BASE = (USER_BASE + 0x1040)
    USER_TI_VPE_BASE = (USER_BASE + 0x1050)
    USER_SAA7134_BASE = (USER_BASE + 0x1060)
    USER_ADV7180_BASE = (USER_BASE + 0x1070)
    USER_TC358743_BASE = (USER_BASE + 0x1080)

    MPEG_BASE = (V4L2_CTRL_CLASS.MPEG | 0x900)
    MPEG_CLASS = (V4L2_CTRL_CLASS.MPEG | 1)

    MPEG_STREAM_TYPE = (MPEG_BASE+0)
    MPEG_STREAM_PID_PMT = (MPEG_BASE+1)
    MPEG_STREAM_PID_AUDIO = (MPEG_BASE+2)
    MPEG_STREAM_PID_VIDEO = (MPEG_BASE+3)
    MPEG_STREAM_PID_PCR = (MPEG_BASE+4)
    MPEG_STREAM_PES_ID_AUDIO = (MPEG_BASE+5)
    MPEG_STREAM_PES_ID_VIDEO = (MPEG_BASE+6)
    MPEG_STREAM_VBI_FMT = (MPEG_BASE+7)

    MPEG_AUDIO_SAMPLING_FREQ = (MPEG_BASE+100)
    MPEG_AUDIO_ENCODING = (MPEG_BASE+101)
    MPEG_AUDIO_L1_BITRATE = (MPEG_BASE+102)
    MPEG_AUDIO_L2_BITRATE = (MPEG_BASE+103)
    MPEG_AUDIO_L3_BITRATE = (MPEG_BASE+104)
    MPEG_AUDIO_MODE = (MPEG_BASE+105)
    MPEG_AUDIO_MODE_EXTENSION = (MPEG_BASE+106)
    MPEG_AUDIO_EMPHASIS = (MPEG_BASE+107)
    MPEG_AUDIO_CRC = (MPEG_BASE+108)
    MPEG_AUDIO_MUTE = (MPEG_BASE+109)
    MPEG_AUDIO_AAC_BITRATE = (MPEG_BASE+110)
    MPEG_AUDIO_AC3_BITRATE = (MPEG_BASE+111)
    MPEG_AUDIO_DEC_PLAYBACK = (MPEG_BASE+112)
    MPEG_AUDIO_DEC_MULTILINGUAL_PLAYBACK = (MPEG_BASE+113)

    MPEG_VIDEO_ENCODING = (MPEG_BASE+200)
    MPEG_VIDEO_ASPECT = (MPEG_BASE+201)
    MPEG_VIDEO_B_FRAMES = (MPEG_BASE+202)
    MPEG_VIDEO_GOP_SIZE = (MPEG_BASE+203)
    MPEG_VIDEO_GOP_CLOSURE = (MPEG_BASE+204)
    MPEG_VIDEO_PULLDOWN = (MPEG_BASE+205)
    MPEG_VIDEO_BITRATE_MODE = (MPEG_BASE+206)
    MPEG_VIDEO_BITRATE = (MPEG_BASE+207)
    MPEG_VIDEO_BITRATE_PEAK = (MPEG_BASE+208)
    MPEG_VIDEO_TEMPORAL_DECIMATION = (MPEG_BASE+209)
    MPEG_VIDEO_MUTE = (MPEG_BASE+210)
    MPEG_VIDEO_MUTE_YUV = (MPEG_BASE+211)
    MPEG_VIDEO_DECODER_SLICE_INTERFACE = (MPEG_BASE+212)
    MPEG_VIDEO_DECODER_MPEG4_DEBLOCK_FILTER = (MPEG_BASE+213)
    MPEG_VIDEO_CYCLIC_INTRA_REFRESH_MB = (MPEG_BASE+214)
    MPEG_VIDEO_FRAME_RC_ENABLE = (MPEG_BASE+215)
    MPEG_VIDEO_HEADER_MODE = (MPEG_BASE+216)
    MPEG_VIDEO_MAX_REF_PIC = (MPEG_BASE+217)
    MPEG_VIDEO_MB_RC_ENABLE = (MPEG_BASE+218)
    MPEG_VIDEO_MULTI_SLICE_MAX_BYTES = (MPEG_BASE+219)
    MPEG_VIDEO_MULTI_SLICE_MAX_MB = (MPEG_BASE+220)
    MPEG_VIDEO_MULTI_SLICE_MODE = (MPEG_BASE+221)
    MPEG_VIDEO_VBV_SIZE = (MPEG_BASE+222)
    MPEG_VIDEO_DEC_PTS = (MPEG_BASE+223)
    MPEG_VIDEO_DEC_FRAME = (MPEG_BASE+224)
    MPEG_VIDEO_VBV_DELAY = (MPEG_BASE+225)
    MPEG_VIDEO_REPEAT_SEQ_HEADER = (MPEG_BASE+226)
    MPEG_VIDEO_MV_H_SEARCH_RANGE = (MPEG_BASE+227)
    MPEG_VIDEO_MV_V_SEARCH_RANGE = (MPEG_BASE+228)

    MPEG_VIDEO_FORCE_KEY_FRAME = (MPEG_BASE+229)
    MPEG_VIDEO_H263_I_FRAME_QP = (MPEG_BASE+300)
    MPEG_VIDEO_H263_P_FRAME_QP = (MPEG_BASE+301)
    MPEG_VIDEO_H263_B_FRAME_QP = (MPEG_BASE+302)
    MPEG_VIDEO_H263_MIN_QP = (MPEG_BASE+303)
    MPEG_VIDEO_H263_MAX_QP = (MPEG_BASE+304)
    MPEG_VIDEO_H264_I_FRAME_QP = (MPEG_BASE+350)
    MPEG_VIDEO_H264_P_FRAME_QP = (MPEG_BASE+351)
    MPEG_VIDEO_H264_B_FRAME_QP = (MPEG_BASE+352)
    MPEG_VIDEO_H264_MIN_QP = (MPEG_BASE+353)
    MPEG_VIDEO_H264_MAX_QP = (MPEG_BASE+354)
    MPEG_VIDEO_H264_8X8_TRANSFORM = (MPEG_BASE+355)
    MPEG_VIDEO_H264_CPB_SIZE = (MPEG_BASE+356)
    MPEG_VIDEO_H264_ENTROPY_MODE = (MPEG_BASE+357)
    MPEG_VIDEO_H264_I_PERIOD = (MPEG_BASE+358)
    MPEG_VIDEO_H264_LEVEL = (MPEG_BASE+359)
    MPEG_VIDEO_H264_LOOP_FILTER_ALPHA = (MPEG_BASE+360)
    MPEG_VIDEO_H264_LOOP_FILTER_BETA = (MPEG_BASE+361)
    MPEG_VIDEO_H264_LOOP_FILTER_MODE = (MPEG_BASE+362)
    MPEG_VIDEO_H264_PROFILE = (MPEG_BASE+363)
    MPEG_VIDEO_H264_VUI_EXT_SAR_HEIGHT = (MPEG_BASE+364)
    MPEG_VIDEO_H264_VUI_EXT_SAR_WIDTH = (MPEG_BASE+365)
    MPEG_VIDEO_H264_VUI_SAR_ENABLE = (MPEG_BASE+366)
    MPEG_VIDEO_H264_VUI_SAR_IDC = (MPEG_BASE+367)
    MPEG_VIDEO_H264_SEI_FRAME_PACKING = (MPEG_BASE+368)
    MPEG_VIDEO_H264_SEI_FP_CURRENT_FRAME_0 = (MPEG_BASE+369)
    MPEG_VIDEO_H264_SEI_FP_ARRANGEMENT_TYPE = (MPEG_BASE+370)
    MPEG_VIDEO_H264_FMO = (MPEG_BASE+371)
    MPEG_VIDEO_H264_FMO_MAP_TYPE = (MPEG_BASE+372)
    MPEG_VIDEO_H264_FMO_SLICE_GROUP = (MPEG_BASE+373)
    MPEG_VIDEO_H264_FMO_CHANGE_DIRECTION = (MPEG_BASE+374)
    MPEG_VIDEO_H264_FMO_CHANGE_RATE = (MPEG_BASE+375)
    MPEG_VIDEO_H264_FMO_RUN_LENGTH = (MPEG_BASE+376)
    MPEG_VIDEO_H264_ASO = (MPEG_BASE+377)
    MPEG_VIDEO_H264_ASO_SLICE_ORDER = (MPEG_BASE+378)
    MPEG_VIDEO_H264_HIERARCHICAL_CODING = (MPEG_BASE+379)
    MPEG_VIDEO_H264_HIERARCHICAL_CODING_TYPE = (MPEG_BASE+380)
    MPEG_VIDEO_H264_HIERARCHICAL_CODING_LAYER = (MPEG_BASE+381)
    MPEG_VIDEO_H264_HIERARCHICAL_CODING_LAYER_QP = (MPEG_BASE+382)

    MPEG_VIDEO_MPEG4_I_FRAME_QP = (MPEG_BASE+400)
    MPEG_VIDEO_MPEG4_P_FRAME_QP = (MPEG_BASE+401)
    MPEG_VIDEO_MPEG4_B_FRAME_QP = (MPEG_BASE+402)
    MPEG_VIDEO_MPEG4_MIN_QP = (MPEG_BASE+403)
    MPEG_VIDEO_MPEG4_MAX_QP = (MPEG_BASE+404)
    MPEG_VIDEO_MPEG4_LEVEL = (MPEG_BASE+405)
    MPEG_VIDEO_MPEG4_PROFILE = (MPEG_BASE+406)
    MPEG_VIDEO_MPEG4_QPEL = (MPEG_BASE+407)

    MPEG_VIDEO_VPX_NUM_PARTITIONS = (MPEG_BASE+500)
    MPEG_VIDEO_VPX_IMD_DISABLE_4X4 = (MPEG_BASE+501)
    MPEG_VIDEO_VPX_NUM_REF_FRAMES = (MPEG_BASE+502)
    MPEG_VIDEO_VPX_FILTER_LEVEL = (MPEG_BASE+503)
    MPEG_VIDEO_VPX_FILTER_SHARPNESS = (MPEG_BASE+504)
    MPEG_VIDEO_VPX_GOLDEN_FRAME_REF_PERIOD = (MPEG_BASE+505)
    MPEG_VIDEO_VPX_GOLDEN_FRAME_SEL = (MPEG_BASE+506)
    MPEG_VIDEO_VPX_MIN_QP = (MPEG_BASE+507)
    MPEG_VIDEO_VPX_MAX_QP = (MPEG_BASE+508)
    MPEG_VIDEO_VPX_I_FRAME_QP = (MPEG_BASE+509)
    MPEG_VIDEO_VPX_P_FRAME_QP = (MPEG_BASE+510)
    MPEG_VIDEO_VPX_PROFILE = (MPEG_BASE+511)

    MPEG_CX2341X_BASE = (V4L2_CTRL_CLASS.MPEG | 0x1000)
    MPEG_CX2341X_VIDEO_SPATIAL_FILTER_MODE = (MPEG_CX2341X_BASE+0)
    MPEG_CX2341X_VIDEO_SPATIAL_FILTER = (MPEG_CX2341X_BASE+1)
    MPEG_CX2341X_VIDEO_LUMA_SPATIAL_FILTER_TYPE = (MPEG_CX2341X_BASE+2)
    MPEG_CX2341X_VIDEO_CHROMA_SPATIAL_FILTER_TYPE = (MPEG_CX2341X_BASE+3)
    MPEG_CX2341X_VIDEO_TEMPORAL_FILTER_MODE = (MPEG_CX2341X_BASE+4)
    MPEG_CX2341X_VIDEO_TEMPORAL_FILTER = (MPEG_CX2341X_BASE+5)
    MPEG_CX2341X_VIDEO_MEDIAN_FILTER_TYPE = (MPEG_CX2341X_BASE+6)
    MPEG_CX2341X_VIDEO_LUMA_MEDIAN_FILTER_BOTTOM = (MPEG_CX2341X_BASE+7)
    MPEG_CX2341X_VIDEO_LUMA_MEDIAN_FILTER_TOP = (MPEG_CX2341X_BASE+8)
    MPEG_CX2341X_VIDEO_CHROMA_MEDIAN_FILTER_BOTTOM = (MPEG_CX2341X_BASE+9)
    MPEG_CX2341X_VIDEO_CHROMA_MEDIAN_FILTER_TOP = (MPEG_CX2341X_BASE+10)
    MPEG_CX2341X_STREAM_INSERT_NAV_PACKETS = (MPEG_CX2341X_BASE+11)

    MPEG_MFC51_BASE = (V4L2_CTRL_CLASS.MPEG | 0x1100)
    MPEG_MFC51_VIDEO_DECODER_H264_DISPLAY_DELAY = (MPEG_MFC51_BASE+0)
    MPEG_MFC51_VIDEO_DECODER_H264_DISPLAY_DELAY_ENABLE = (MPEG_MFC51_BASE+1)
    MPEG_MFC51_VIDEO_FRAME_SKIP_MODE = (MPEG_MFC51_BASE+2)
    MPEG_MFC51_VIDEO_FORCE_FRAME_TYPE = (MPEG_MFC51_BASE+3)
    MPEG_MFC51_VIDEO_PADDING = (MPEG_MFC51_BASE+4)
    MPEG_MFC51_VIDEO_PADDING_YUV = (MPEG_MFC51_BASE+5)
    MPEG_MFC51_VIDEO_RC_FIXED_TARGET_BIT = (MPEG_MFC51_BASE+6)
    MPEG_MFC51_VIDEO_RC_REACTION_COEFF = (MPEG_MFC51_BASE+7)
    MPEG_MFC51_VIDEO_H264_ADAPTIVE_RC_ACTIVITY = (MPEG_MFC51_BASE+50)
    MPEG_MFC51_VIDEO_H264_ADAPTIVE_RC_DARK = (MPEG_MFC51_BASE+51)
    MPEG_MFC51_VIDEO_H264_ADAPTIVE_RC_SMOOTH = (MPEG_MFC51_BASE+52)
    MPEG_MFC51_VIDEO_H264_ADAPTIVE_RC_STATIC = (MPEG_MFC51_BASE+53)
    MPEG_MFC51_VIDEO_H264_NUM_REF_PIC_FOR_P = (MPEG_MFC51_BASE+54)

    CAMERA_CLASS_BASE = (V4L2_CTRL_CLASS.CAMERA | 0x900)
    CAMERA_CLASS = (V4L2_CTRL_CLASS.CAMERA | 1)

    EXPOSURE_AUTO = (CAMERA_CLASS_BASE+1)
    EXPOSURE_ABSOLUTE = (CAMERA_CLASS_BASE+2)
    EXPOSURE_AUTO_PRIORITY = (CAMERA_CLASS_BASE+3)

    PAN_RELATIVE = (CAMERA_CLASS_BASE+4)
    TILT_RELATIVE = (CAMERA_CLASS_BASE+5)
    PAN_RESET = (CAMERA_CLASS_BASE+6)
    TILT_RESET = (CAMERA_CLASS_BASE+7)
    PAN_ABSOLUTE = (CAMERA_CLASS_BASE+8)
    TILT_ABSOLUTE = (CAMERA_CLASS_BASE+9)

    FOCUS_ABSOLUTE = (CAMERA_CLASS_BASE+10)
    FOCUS_RELATIVE = (CAMERA_CLASS_BASE+11)
    FOCUS_AUTO = (CAMERA_CLASS_BASE+12)

    ZOOM_ABSOLUTE = (CAMERA_CLASS_BASE+13)
    ZOOM_RELATIVE = (CAMERA_CLASS_BASE+14)
    ZOOM_CONTINUOUS = (CAMERA_CLASS_BASE+15)

    PRIVACY = (CAMERA_CLASS_BASE+16)

    IRIS_ABSOLUTE = (CAMERA_CLASS_BASE+17)
    IRIS_RELATIVE = (CAMERA_CLASS_BASE+18)

    AUTO_EXPOSURE_BIAS = (CAMERA_CLASS_BASE+19)
    AUTO_N_PRESET_WHITE_BALANCE = (CAMERA_CLASS_BASE+20)

    WIDE_DYNAMIC_RANGE = (CAMERA_CLASS_BASE+21)
    IMAGE_STABILIZATION = (CAMERA_CLASS_BASE+22)

    ISO_SENSITIVITY = (CAMERA_CLASS_BASE+23)
    ISO_SENSITIVITY_AUTO = (CAMERA_CLASS_BASE+24)
    EXPOSURE_METERING = (CAMERA_CLASS_BASE+25)

    SCENE_MODE = (CAMERA_CLASS_BASE+26)

    BISMASK_3A_LOCK = (CAMERA_CLASS_BASE+27)

    AUTO_FOCUS_START = (CAMERA_CLASS_BASE+28)
    AUTO_FOCUS_STOP = (CAMERA_CLASS_BASE+29)
    AUTO_FOCUS_STATUS = (CAMERA_CLASS_BASE+30)

    AUTO_FOCUS_RANGE = (CAMERA_CLASS_BASE+31)

    PAN_SPEED = (CAMERA_CLASS_BASE+32)
    TILT_SPEED = (CAMERA_CLASS_BASE+33)

    FM_TX_CLASS_BASE = (V4L2_CTRL_CLASS.FM_TX | 0x900)
    FM_TX_CLASS = (V4L2_CTRL_CLASS.FM_TX | 1)

    RDS_TX_DEVIATION = (FM_TX_CLASS_BASE + 1)
    RDS_TX_PI = (FM_TX_CLASS_BASE + 2)
    RDS_TX_PTY = (FM_TX_CLASS_BASE + 3)
    RDS_TX_PS_NAME = (FM_TX_CLASS_BASE + 5)
    RDS_TX_RADIO_TEXT = (FM_TX_CLASS_BASE + 6)
    RDS_TX_MONO_STEREO = (FM_TX_CLASS_BASE + 7)
    RDS_TX_ARTIFICIAL_HEAD = (FM_TX_CLASS_BASE + 8)
    RDS_TX_COMPRESSED = (FM_TX_CLASS_BASE + 9)
    RDS_TX_DYNAMIC_PTY = (FM_TX_CLASS_BASE + 10)
    RDS_TX_TRAFFIC_ANNOUNCEMENT = (FM_TX_CLASS_BASE + 11)
    RDS_TX_TRAFFIC_PROGRAM = (FM_TX_CLASS_BASE + 12)
    RDS_TX_MUSIC_SPEECH = (FM_TX_CLASS_BASE + 13)
    RDS_TX_ALT_FREQS_ENABLE = (FM_TX_CLASS_BASE + 14)
    RDS_TX_ALT_FREQS = (FM_TX_CLASS_BASE + 15)

    AUDIO_LIMITER_ENABLED = (FM_TX_CLASS_BASE + 64)
    AUDIO_LIMITER_RELEASE_TIME = (FM_TX_CLASS_BASE + 65)
    AUDIO_LIMITER_DEVIATION = (FM_TX_CLASS_BASE + 66)

    AUDIO_COMPRESSION_ENABLED = (FM_TX_CLASS_BASE + 80)
    AUDIO_COMPRESSION_GAIN = (FM_TX_CLASS_BASE + 81)
    AUDIO_COMPRESSION_THRESHOLD = (FM_TX_CLASS_BASE + 82)
    AUDIO_COMPRESSION_ATTACK_TIME = (FM_TX_CLASS_BASE + 83)
    AUDIO_COMPRESSION_RELEASE_TIME = (FM_TX_CLASS_BASE + 84)

    PILOT_TONE_ENABLED = (FM_TX_CLASS_BASE + 96)
    PILOT_TONE_DEVIATION = (FM_TX_CLASS_BASE + 97)
    PILOT_TONE_FREQUENCY = (FM_TX_CLASS_BASE + 98)

    TUNE_PREEMPHASIS = (FM_TX_CLASS_BASE + 112)
    TUNE_POWER_LEVEL = (FM_TX_CLASS_BASE + 113)
    TUNE_ANTENNA_CAPACITOR = (FM_TX_CLASS_BASE + 114)

    FLASH_CLASS_BASE = (V4L2_CTRL_CLASS.FLASH | 0x900)
    FLASH_CLASS = (V4L2_CTRL_CLASS.FLASH | 1)

    FLASH_LED_MODE = (FLASH_CLASS_BASE + 1)
    FLASH_STROBE_SOURCE = (FLASH_CLASS_BASE + 2)

    FLASH_STROBE = (FLASH_CLASS_BASE + 3)
    FLASH_STROBE_STOP = (FLASH_CLASS_BASE + 4)
    FLASH_STROBE_STATUS = (FLASH_CLASS_BASE + 5)
    FLASH_TIMEOUT = (FLASH_CLASS_BASE + 6)
    FLASH_INTENSITY = (FLASH_CLASS_BASE + 7)
    FLASH_TORCH_INTENSITY = (FLASH_CLASS_BASE + 8)
    FLASH_INDICATOR_INTENSITY = (FLASH_CLASS_BASE + 9)
    FLASH_FAULT = (FLASH_CLASS_BASE + 10)
    FLASH_CHARGE = (FLASH_CLASS_BASE + 11)
    FLASH_READY = (FLASH_CLASS_BASE + 12)

    JPEG_CLASS_BASE = (V4L2_CTRL_CLASS.JPEG | 0x900)
    JPEG_CLASS = (V4L2_CTRL_CLASS.JPEG | 1)

    JPEG_CHROMA_SUBSAMPLING = (JPEG_CLASS_BASE + 1)
    JPEG_RESTART_INTERVAL = (JPEG_CLASS_BASE + 2)
    JPEG_COMPRESSION_QUALITY = (JPEG_CLASS_BASE + 3)

    JPEG_ACTIVE_MARKER = (JPEG_CLASS_BASE + 4)

    IMAGE_SOURCE_CLASS_BASE = (V4L2_CTRL_CLASS.IMAGE_SOURCE | 0x900)
    IMAGE_SOURCE_CLASS = (V4L2_CTRL_CLASS.IMAGE_SOURCE | 1)

    VBLANK = (IMAGE_SOURCE_CLASS_BASE + 1)
    HBLANK = (IMAGE_SOURCE_CLASS_BASE + 2)
    ANALOGUE_GAIN = (IMAGE_SOURCE_CLASS_BASE + 3)

    TEST_PATTERN_RED = (IMAGE_SOURCE_CLASS_BASE + 4)
    TEST_PATTERN_GREENR = (IMAGE_SOURCE_CLASS_BASE + 5)
    TEST_PATTERN_BLUE = (IMAGE_SOURCE_CLASS_BASE + 6)
    TEST_PATTERN_GREENB = (IMAGE_SOURCE_CLASS_BASE + 7)

    IMAGE_PROC_CLASS_BASE = (V4L2_CTRL_CLASS.IMAGE_PROC | 0x900)
    IMAGE_PROC_CLASS = (V4L2_CTRL_CLASS.IMAGE_PROC | 1)

    LINK_FREQ = (IMAGE_PROC_CLASS_BASE + 1)
    PIXEL_RATE = (IMAGE_PROC_CLASS_BASE + 2)
    TEST_PATTERN = (IMAGE_PROC_CLASS_BASE + 3)

    DV_CLASS_BASE = (V4L2_CTRL_CLASS.DV | 0x900)
    DV_CLASS = (V4L2_CTRL_CLASS.DV | 1)

    DV_TX_HOTPLUG = (DV_CLASS_BASE + 1)
    DV_TX_RXSENSE = (DV_CLASS_BASE + 2)
    DV_TX_EDID_PRESENT = (DV_CLASS_BASE + 3)
    DV_TX_MODE = (DV_CLASS_BASE + 4)
    DV_TX_RGB_RANGE = (DV_CLASS_BASE + 5)
    DV_TX_IT_CONTENT_TYPE = (DV_CLASS_BASE + 6)

    DV_RX_POWER_PRESENT = (DV_CLASS_BASE + 100)
    DV_RX_RGB_RANGE = (DV_CLASS_BASE + 101)
    DV_RX_IT_CONTENT_TYPE = (DV_CLASS_BASE + 102)

    FM_RX_CLASS_BASE = (V4L2_CTRL_CLASS.FM_RX | 0x900)
    FM_RX_CLASS = (V4L2_CTRL_CLASS.FM_RX | 1)

    TUNE_DEEMPHASIS = (FM_RX_CLASS_BASE + 1)

    RDS_RECEPTION = (FM_RX_CLASS_BASE + 2)
    RDS_RX_PTY = (FM_RX_CLASS_BASE + 3)
    RDS_RX_PS_NAME = (FM_RX_CLASS_BASE + 4)
    RDS_RX_RADIO_TEXT = (FM_RX_CLASS_BASE + 5)
    RDS_RX_TRAFFIC_ANNOUNCEMENT = (FM_RX_CLASS_BASE + 6)
    RDS_RX_TRAFFIC_PROGRAM = (FM_RX_CLASS_BASE + 7)
    RDS_RX_MUSIC_SPEECH = (FM_RX_CLASS_BASE + 8)

    RF_TUNER_CLASS_BASE = (V4L2_CTRL_CLASS.RF_TUNER | 0x900)
    RF_TUNER_CLASS = (V4L2_CTRL_CLASS.RF_TUNER | 1)

    RF_TUNER_BANDWIDTH_AUTO = (RF_TUNER_CLASS_BASE + 11)
    RF_TUNER_BANDWIDTH = (RF_TUNER_CLASS_BASE + 12)
    RF_TUNER_RF_GAIN = (RF_TUNER_CLASS_BASE + 32)
    RF_TUNER_LNA_GAIN_AUTO = (RF_TUNER_CLASS_BASE + 41)
    RF_TUNER_LNA_GAIN = (RF_TUNER_CLASS_BASE + 42)
    RF_TUNER_MIXER_GAIN_AUTO = (RF_TUNER_CLASS_BASE + 51)
    RF_TUNER_MIXER_GAIN = (RF_TUNER_CLASS_BASE + 52)
    RF_TUNER_IF_GAIN_AUTO = (RF_TUNER_CLASS_BASE + 61)
    RF_TUNER_IF_GAIN = (RF_TUNER_CLASS_BASE + 62)
    RF_TUNER_PLL_LOCK = (RF_TUNER_CLASS_BASE + 91)

    DETECT_CLASS_BASE = (V4L2_CTRL_CLASS.DETECT | 0x900)
    DETECT_CLASS = (V4L2_CTRL_CLASS.DETECT | 1)

    DETECT_MD_MODE = (DETECT_CLASS_BASE + 1)
    DETECT_MD_GLOBAL_THRESHOLD = (DETECT_CLASS_BASE + 2)
    DETECT_MD_THRESHOLD_GRID = (DETECT_CLASS_BASE + 3)
    DETECT_MD_REGION_GRID = (DETECT_CLASS_BASE + 4)


class V4L2_POWER_LINE(enum.IntEnum):
    FREQUENCY_DISABLED = 0
    FREQUENCY_50HZ = 1
    FREQUENCY_60HZ = 2
    FREQUENCY_AUTO = 3


class V4L2_COLORFX(enum.IntEnum):
    NONE = 0
    BW = 1
    SEPIA = 2
    NEGATIVE = 3
    EMBOSS = 4
    SKETCH = 5
    SKY_BLUE = 6
    GRASS_GREEN = 7
    SKIN_WHITEN = 8
    VIVID = 9
    AQUA = 10
    ART_FREEZE = 11
    SILHOUETTE = 12
    SOLARIZATION = 13
    ANTIQUE = 14
    SET_CBCR = 15


class V4L2_MPEG_STREAM_TYPE(enum.IntEnum):
    MPEG2_PS = 0
    MPEG2_TS = 1
    MPEG1_SS = 2
    MPEG2_DVD = 3
    MPEG1_VCD = 4
    MPEG2_SVCD = 5


class V4L2_MPEG_STREAM_VBI_FMT(enum.IntEnum):
    NONE = 0
    IVTV = 1


class V4L2_MPEG_AUDIO_SAMPLING(enum.IntEnum):
    FREQ_44100 = 0
    FREQ_48000 = 1
    FREQ_32000 = 2


class V4L2_MPEG_AUDIO_ENCODING(enum.IntEnum):
    LAYER_1 = 0
    LAYER_2 = 1
    LAYER_3 = 2
    AAC = 3
    AC3 = 4


class V4L2_MPEG_AUDIO_L1(enum.IntEnum):
    BITRATE_32K = 0
    BITRATE_64K = 1
    BITRATE_96K = 2
    BITRATE_128K = 3
    BITRATE_160K = 4
    BITRATE_192K = 5
    BITRATE_224K = 6
    BITRATE_256K = 7
    BITRATE_288K = 8
    BITRATE_320K = 9
    BITRATE_352K = 10
    BITRATE_384K = 11
    BITRATE_416K = 12
    BITRATE_448K = 13


class V4L2_MPEG_AUDIO_L2(enum.IntEnum):
    BITRATE_32K = 0
    BITRATE_48K = 1
    BITRATE_56K = 2
    BITRATE_64K = 3
    BITRATE_80K = 4
    BITRATE_96K = 5
    BITRATE_112K = 6
    BITRATE_128K = 7
    BITRATE_160K = 8
    BITRATE_192K = 9
    BITRATE_224K = 10
    BITRATE_256K = 11
    BITRATE_320K = 12
    BITRATE_384K = 13


class V4L2_MPEG_AUDIO_L3(enum.IntEnum):
    BITRATE_32K = 0
    BITRATE_40K = 1
    BITRATE_48K = 2
    BITRATE_56K = 3
    BITRATE_64K = 4
    BITRATE_80K = 5
    BITRATE_96K = 6
    BITRATE_112K = 7
    BITRATE_128K = 8
    BITRATE_160K = 9
    BITRATE_192K = 10
    BITRATE_224K = 11
    BITRATE_256K = 12
    BITRATE_320K = 13


class V4L2_MPEG_AUDIO_MODE(enum.IntEnum):
    STEREO = 0
    JOINT_STEREO = 1
    DUAL = 2
    MONO = 3


class V4L2_MPEG_AUDIO_MODE_EXTENSION(enum.IntEnum):
    BOUND_4 = 0
    BOUND_8 = 1
    BOUND_12 = 2
    BOUND_16 = 3


class V4L2_MPEG_AUDIO(enum.IntEnum):
    EMPHASIS_NONE = 0
    EMPHASIS_50_DIV_15_uS = 1
    EMPHASIS_CCITT_J17 = 2


class V4L2_MPEG_AUDIO_CRC(enum.IntEnum):
    NONE = 0
    CRC16 = 1


class V4L2_MPEG_AUDIO_AC3(enum.IntEnum):
    BITRATE_32K = 0
    BITRATE_40K = 1
    BITRATE_48K = 2
    BITRATE_56K = 3
    BITRATE_64K = 4
    BITRATE_80K = 5
    BITRATE_96K = 6
    BITRATE_112K = 7
    BITRATE_128K = 8
    BITRATE_160K = 9
    BITRATE_192K = 10
    BITRATE_224K = 11
    BITRATE_256K = 12
    BITRATE_320K = 13
    BITRATE_384K = 14
    BITRATE_448K = 15
    BITRATE_512K = 16
    BITRATE_576K = 17
    BITRATE_640K = 18


class V4L2_MPEG_AUDIO_DEC_PLAYBACK(enum.IntEnum):
    AUTO = 0
    STEREO = 1
    LEFT = 2
    RIGHT = 3
    MONO = 4
    SWAPPED_STEREO = 5


class V4L2_MPEG_VIDEO_ENCODING(enum.IntEnum):
    MPEG_1 = 0
    MPEG_2 = 1
    MPEG_4_AVC = 2


class V4L2_MPEG_VIDEO(enum.IntEnum):
    ASPECT_1x1 = 0
    ASPECT_4x3 = 1
    ASPECT_16x9 = 2
    ASPECT_221x100 = 3


class V4L2_MPEG_VIDEO_BITRATE_MODE(enum.IntEnum):
    VBR = 0
    CBR = 1


class V4L2_MPEG_VIDEO_HEADER_MODE(enum.IntEnum):
    SEPARATE = 0
    JOINED_WITH_1ST_FRAME = 1


class V4L2_MPEG_VIDEO_MULTI_SLICE_MODE(enum.IntEnum):
    SINGLE = 0
    MAX_MB = 1
    MAX_BYTES = 2


class V4L2_MPEG_VIDEO_H264_ENTROPY_MODE(enum.IntEnum):
    CAVLC = 0
    CABAC = 1


class V4L2_MPEG_VIDEO_H264(enum.IntEnum):
    LEVEL_1_0 = 0
    LEVEL_1B = 1
    LEVEL_1_1 = 2
    LEVEL_1_2 = 3
    LEVEL_1_3 = 4
    LEVEL_2_0 = 5
    LEVEL_2_1 = 6
    LEVEL_2_2 = 7
    LEVEL_3_0 = 8
    LEVEL_3_1 = 9
    LEVEL_3_2 = 10
    LEVEL_4_0 = 11
    LEVEL_4_1 = 12
    LEVEL_4_2 = 13
    LEVEL_5_0 = 14
    LEVEL_5_1 = 15


class V4L2_MPEG_VIDEO_H264_LOOP_FILTER_MODE(enum.IntEnum):
    ENABLED = 0
    DISABLED = 1
    DISABLED_AT_SLICE_BOUNDARY = 2


class V4L2_MPEG_VIDEO_H264_PROFILE(enum.IntEnum):
    BASELINE = 0
    CONSTRAINED_BASELINE = 1
    MAIN = 2
    EXTENDED = 3
    HIGH = 4
    HIGH_10 = 5
    HIGH_422 = 6
    HIGH_444_PREDICTIVE = 7
    HIGH_10_INTRA = 8
    HIGH_422_INTRA = 9
    HIGH_444_INTRA = 10
    CAVLC_444_INTRA = 11
    SCALABLE_BASELINE = 12
    SCALABLE_HIGH = 13
    SCALABLE_HIGH_INTRA = 14
    STEREO_HIGH = 15
    MULTIVIEW_HIGH = 16


class V4L2_MPEG_VIDEO_H264_VUI_SAR(enum.IntEnum):
    IDC_UNSPECIFIED = 0
    IDC_1x1 = 1
    IDC_12x11 = 2
    IDC_10x11 = 3
    IDC_16x11 = 4
    IDC_40x33 = 5
    IDC_24x11 = 6
    IDC_20x11 = 7
    IDC_32x11 = 8
    IDC_80x33 = 9
    IDC_18x11 = 10
    IDC_15x11 = 11
    IDC_64x33 = 12
    IDC_160x99 = 13
    IDC_4x3 = 14
    IDC_3x2 = 15
    IDC_2x1 = 16
    IDC_EXTENDED = 17


class V4L2_MPEG_VIDEO_H264_SEI_FP_ARRANGEMENT_TYPE(enum.IntEnum):
    CHECKERBOARD = 0
    COLUMN = 1
    ROW = 2
    SIDE_BY_SIDE = 3
    TOP_BOTTOM = 4
    TEMPORAL = 5


class V4L2_MPEG_VIDEO_H264_FMO_MAP_TYPE(enum.IntEnum):
    INTERLEAVED_SLICES = 0
    SCATTERED_SLICES = 1
    FOREGROUND_WITH_LEFT_OVER = 2
    BOX_OUT = 3
    RASTER_SCAN = 4
    WIPE_SCAN = 5
    EXPLICIT = 6


class V4L2_MPEG_VIDEO_H264_FMO_CHANGE_DIR(enum.IntEnum):
    RIGHT = 0
    LEFT = 1


class V4L2_MPEG_VIDEO_H264_HIERARCHICAL_CODING(enum.IntEnum):
    B = 0
    P = 1


class V4L2_MPEG_VIDEO_MPEG4(enum.IntEnum):
    LEVEL_0 = 0
    LEVEL_0B = 1
    LEVEL_1 = 2
    LEVEL_2 = 3
    LEVEL_3 = 4
    LEVEL_3B = 5
    LEVEL_4 = 6
    LEVEL_5 = 7


class V4L2_MPEG_VIDEO_MPEG4_PROFILE(enum.IntEnum):
    SIMPLE = 0
    ADVANCED_SIMPLE = 1
    CORE = 2
    SIMPLE_SCALABLE = 3
    ADVANCED_CODING_EFFICIENCY = 4


class V4L2_MPEG_VIDEO_VPX_PARTITIONS(enum.IntEnum):
    NUM_1 = 0
    NUM_2 = 1
    NUM_4 = 2
    NUM_8 = 3


class V4L2_MPEG_VIDEO_VPX_REF_FRAME(enum.IntEnum):
    NUM_1 = 0
    NUM_2 = 1
    NUM_3 = 2


class V4L2_MPEG_VIDEO_VPX_GOLDEN_FRAME_USE(enum.IntEnum):
    PREV = 0
    REF_PERIOD = 1


class V4L2_MPEG_CX2341X_VIDEO_SPATIAL_FILTER_MODE(enum.IntEnum):
    MANUAL = 0
    AUTO = 1


class V4L2_MPEG_CX2341X_VIDEO_LUMA_SPATIAL_FILTER(enum.IntEnum):
    TYPE_OFF = 0
    TYPE_1D_HOR = 1
    TYPE_1D_VERT = 2
    TYPE_2D_HV_SEPARABLE = 3
    TYPE_2D_SYM_NON_SEPARABLE = 4


class V4L2_MPEG_CX2341X_VIDEO_CHROMA_SPATIAL_FILTER(enum.IntEnum):
    TYPE_OFF = 0
    TYPE_1D_HOR = 1


class V4L2_MPEG_CX2341X_VIDEO_TEMPORAL_FILTER_MODE(enum.IntEnum):
    MANUAL = 0
    AUTO = 1


class V4L2_MPEG_CX2341X_VIDEO_MEDIAN_FILTER_TYPE(enum.IntEnum):
    OFF = 0
    HOR = 1
    VERT = 2
    HOR_VERT = 3
    DIAG = 4


class V4L2_MPEG_MFC51_VIDEO_FRAME_SKIP_MODE(enum.IntEnum):
    DISABLED = 0
    LEVEL_LIMIT = 1
    BUF_LIMIT = 2


class V4L2_MPEG_MFC51_VIDEO_FORCE_FRAME_TYPE(enum.IntEnum):
    DISABLED = 0
    I_FRAME = 1
    NOT_CODED = 2


class V4L2_EXPOSURE(enum.IntEnum):
    AUTO = 0
    MANUAL = 1
    SHUTTER_PRIORITY = 2
    APERTURE_PRIORITY = 3


class V4L2_WHITE_BALANCE(enum.IntEnum):
    MANUAL = 0
    AUTO = 1
    INCANDESCENT = 2
    FLUORESCENT = 3
    FLUORESCENT_H = 4
    HORIZON = 5
    DAYLIGHT = 6
    FLASH = 7
    CLOUDY = 8
    SHADE = 9


class V4L2_ISO_SENSITIVITY(enum.IntEnum):
    MANUAL = 0
    AUTO = 1


class V4L2_EXPOSURE_METERING(enum.IntEnum):
    AVERAGE = 0
    CENTER_WEIGHTED = 1
    SPOT = 2
    MATRIX = 3


class V4L2_SCENE_MODE(enum.IntEnum):
    NONE = 0
    BACKLIGHT = 1
    BEACH_SNOW = 2
    CANDLE_LIGHT = 3
    DAWN_DUSK = 4
    FALL_COLORS = 5
    FIREWORKS = 6
    LANDSCAPE = 7
    NIGHT = 8
    PARTY_INDOOR = 9
    PORTRAIT = 10
    SPORTS = 11
    SUNSET = 12
    TEXT = 13


class V4L2_AUTO_FOCUS_RANGE(enum.IntEnum):
    AUTO = 0
    NORMAL = 1
    MACRO = 2
    INFINITY = 3


class V4L2_PREEMPHASIS(enum.IntEnum):
    DISABLED = 0
    TUNE_50_uS = 1
    TUNE_75_uS = 2


class V4L2_FLASH_LED_MODE(enum.IntEnum):
    NONE = 0
    FLASH = 1
    TORCH = 2


class V4L2_FLASH_STROBE_SOURCE(enum.IntEnum):
    SOFTWARE = 0
    EXTERNAL = 1


class V4L2_JPEG_CHROMA(enum.IntEnum):
    SUBSAMPLING_444 = 0
    SUBSAMPLING_422 = 1
    SUBSAMPLING_420 = 2
    SUBSAMPLING_411 = 3
    SUBSAMPLING_410 = 4
    SUBSAMPLING_GRAY = 5


class V4L2_DV_TX_MODE(enum.IntEnum):
    DVI_D = 0
    HDMI = 1


class V4L2_DV_RGB_RANGE(enum.IntEnum):
    AUTO = 0
    LIMITED = 1
    FULL = 2


class V4L2_DV_IT_CONTENT_TYPE(enum.IntEnum):
    GRAPHICS = 0
    PHOTO = 1
    CINEMA = 2
    GAME = 3
    NO_ITC = 4


class V4L2_DEEMPHASIS(enum.IntEnum):
    DISABLED = V4L2_PREEMPHASIS.DISABLED
    TUNE_50_uS = V4L2_PREEMPHASIS.TUNE_50_uS
    TUNE_75_uS = V4L2_PREEMPHASIS.TUNE_75_uS


class V4L2_DETECT_MD_MODE(enum.IntEnum):
    DISABLED = 0
    GLOBAL = 1
    THRESHOLD_GRID = 2
    REGION_GRID = 3


V4L2_LOCK_EXPOSURE = (1 << 0)
V4L2_LOCK_WHITE_BALANCE = (1 << 1)
V4L2_LOCK_FOCUS = (1 << 2)

V4L2_AUTO_FOCUS_STATUS_IDLE = (0 << 0)
V4L2_AUTO_FOCUS_STATUS_BUSY = (1 << 0)
V4L2_AUTO_FOCUS_STATUS_REACHED = (1 << 1)
V4L2_AUTO_FOCUS_STATUS_FAILED = (1 << 2)

V4L2_FLASH_FAULT_OVER_VOLTAGE = (1 << 0)
V4L2_FLASH_FAULT_TIMEOUT = (1 << 1)
V4L2_FLASH_FAULT_OVER_TEMPERATURE = (1 << 2)
V4L2_FLASH_FAULT_SHORT_CIRCUIT = (1 << 3)
V4L2_FLASH_FAULT_OVER_CURRENT = (1 << 4)
V4L2_FLASH_FAULT_INDICATOR = (1 << 5)
V4L2_FLASH_FAULT_UNDER_VOLTAGE = (1 << 6)
V4L2_FLASH_FAULT_INPUT_VOLTAGE = (1 << 7)
V4L2_FLASH_FAULT_LED_OVER_TEMPERATURE = (1 << 8)

V4L2_JPEG_ACTIVE_MARKER_APP0 = (1 << 0)
V4L2_JPEG_ACTIVE_MARKER_APP1 = (1 << 1)
V4L2_JPEG_ACTIVE_MARKER_COM = (1 << 16)
V4L2_JPEG_ACTIVE_MARKER_DQT = (1 << 17)
V4L2_JPEG_ACTIVE_MARKER_DHT = (1 << 18)


class V4L2_CTRL_TYPE(enum.IntEnum):
    INTEGER = 1
    BOOLEAN = 2
    MENU = 3
    BUTTON = 4
    INTEGER64 = 5
    CTRL_CLASS = 6
    STRING = 7
    BITMASK = 8
    INTEGER_MENU = 9
    COMPOUND = 0x0100
    U8 = 0x0100
    U16 = 0x0101
    U32 = 0x0102


V4L2_CTRL_FLAG_DISABLED = 0x0001
V4L2_CTRL_FLAG_GRABBED = 0x0002
V4L2_CTRL_FLAG_READ_ONLY = 0x0004
V4L2_CTRL_FLAG_UPDATE = 0x0008
V4L2_CTRL_FLAG_INACTIVE = 0x0010
V4L2_CTRL_FLAG_SLIDER = 0x0020
V4L2_CTRL_FLAG_WRITE_ONLY = 0x0040
V4L2_CTRL_FLAG_VOLATILE = 0x0080
V4L2_CTRL_FLAG_HAS_PAYLOAD = 0x0100
V4L2_CTRL_FLAG_EXECUTE_ON_WRITE = 0x0200
V4L2_CTRL_FLAG_MODIFY_LAYOUT = 0x0400



import os
import enum
import errno
import itertools
import mmap
import select
from ctypes import *
from ctypes.util import find_library
from actfw_core.v4l2.types import *
from actfw_core.v4l2.control import *
import io


class _libv4l2(object):

    def __init__(self):
        self.lib = None
        path = find_library('v4l2')
        if path is not None:
            self.lib = CDLL(path, use_errno=True)

    def ioctl(self, *args, **kwargs):
        if self.lib is None:
            raise FileNotFoundError("Not found: 'libv4l2.so'")
        return self.lib.v4l2_ioctl(*args, **kwargs)

    def mmap(self, *args, **kwargs):
        if self.lib is None:
            raise FileNotFoundError("Not found: 'libv4l2.so'")
        return self.lib.v4l2_mmap(*args, **kwargs)

    def munmap(self, *args, **kwargs):
        if self.lib is None:
            raise FileNotFoundError("Not found: 'libv4l2.so'")
        return self.lib.v4l2_munmap(*args, **kwargs)


class _libv4lconvert(object):

    def __init__(self):
        self.lib = None
        path = find_library('v4lconvert')
        if path is not None:
            self.lib = CDLL(path, use_errno=True)

    def create(self, *args, **kwargs):
        if self.lib is None:
            raise FileNotFoundError("Not found: 'libv4lconvert.so'")
        return self.lib.v4lconvert_create(*args, **kwargs)

    def convert(self, *args, **kwargs):
        if self.lib is None:
            raise FileNotFoundError("Not found: 'libv4lconvert.so'")
        return self.lib.v4lconvert_convert(*args, **kwargs)

    def try_format(self, *args, **kwargs):
        if self.lib is None:
            raise FileNotFoundError("Not found: 'libv4lconvert.so'")
        return self.lib.v4lconvert_try_format(*args, **kwargs)


_v4l2 = _libv4l2()
_v4lconvert = _libv4lconvert()

_IOC_NRBITS = 8
_IOC_TYPEBITS = 8

_IOC_SIZEBITS = 14
_IOC_DIRBITS = 2

_IOC_NRMASK = (1 << _IOC_NRBITS) - 1
_IOC_TYPEMASK = (1 << _IOC_TYPEBITS) - 1
_IOC_SIZEMASK = (1 << _IOC_SIZEBITS) - 1
_IOC_DIRMASK = (1 << _IOC_DIRBITS) - 1

_IOC_NRSHIFT = 0
_IOC_TYPESHIFT = _IOC_NRSHIFT + _IOC_NRBITS
_IOC_SIZESHIFT = _IOC_TYPESHIFT + _IOC_TYPEBITS
_IOC_DIRSHIFT = _IOC_SIZESHIFT + _IOC_SIZEBITS

_IOC_NONE = 0
_IOC_WRITE = 1
_IOC_READ = 2


def _IOC(dir, type, nr, size):
    return (dir << _IOC_DIRSHIFT) | (ord(type) << _IOC_TYPESHIFT) | (nr << _IOC_NRSHIFT) | (size << _IOC_SIZESHIFT)


def _IO(type, nr): return _IOC(_IOC_NONE, type, nr, 0)


def _IOR(type, nr, size): return _IOC(_IOC_READ, type, nr, sizeof(size))


def _IOW(type, nr, size): return _IOC(_IOC_WRITE, type, nr, sizeof(size))


def _IOWR(type, nr, size): return _IOC(_IOC_READ | _IOC_WRITE, type, nr, sizeof(size))


class _VIDIOC(enum.IntEnum):
    QUERYCAP = _IOR('V', 0, capability)
    ENUM_FMT = _IOWR('V', 2, fmtdesc)
    S_FMT = _IOWR('V', 5, format)
    REQBUFS = _IOWR('V', 8, requestbuffers)
    QUERYBUF = _IOWR('V', 9, buffer)
    QBUF = _IOWR('V', 15, buffer)
    DQBUF = _IOWR('V', 17, buffer)
    STREAMON = _IOW('V', 18, c_int)
    STREAMOFF = _IOW('V', 19, c_int)
    S_PARM = _IOWR('V', 22, streamparm)
    G_CTRL = _IOWR('V', 27, control)
    S_CTRL = _IOWR('V', 28, control)
    QUERYCTRL = _IOWR('V', 36, queryctrl)
    ENUM_FRAMESIZES = _IOWR('V', 74, frmsizeenum)
    ENUM_FRAMEINTERVALS = _IOWR('V', 75, frmivalenum)


_V4L2_CAP_VIDEO_CAPTURE = 0x00000001
_V4L2_CAP_VIDEO_OUTPUT = 0x00000002
_V4L2_CAP_VIDEO_OVERLAY = 0x00000004
_V4L2_CAP_VBI_CAPTURE = 0x00000010
_V4L2_CAP_VBI_OUTPUT = 0x00000020
_V4L2_CAP_SLICED_VBI_CAPTURE = 0x00000040
_V4L2_CAP_SLICED_VBI_OUTPUT = 0x00000080
_V4L2_CAP_RDS_CAPTURE = 0x00000100
_V4L2_CAP_VIDEO_OUTPUT_OVERLAY = 0x00000200
_V4L2_CAP_HW_FREQ_SEEK = 0x00000400
_V4L2_CAP_RDS_OUTPUT = 0x00000800
_V4L2_CAP_VIDEO_CAPTURE_MPLANE = 0x00001000
_V4L2_CAP_VIDEO_OUTPUT_MPLANE = 0x00002000
_V4L2_CAP_VIDEO_M2M_MPLANE = 0x00004000
_V4L2_CAP_VIDEO_M2M = 0x00008000
_V4L2_CAP_TUNER = 0x00010000
_V4L2_CAP_AUDIO = 0x00020000
_V4L2_CAP_RADIO = 0x00040000
_V4L2_CAP_MODULATOR = 0x00080000
_V4L2_CAP_SDR_CAPTURE = 0x00100000
_V4L2_CAP_EXT_PIX_FORMAT = 0x00200000
_V4L2_CAP_SDR_OUTPUT = 0x00400000
_V4L2_CAP_READWRITE = 0x01000000
_V4L2_CAP_ASYNCIO = 0x02000000
_V4L2_CAP_STREAMING = 0x04000000
_V4L2_CAP_TOUCH = 0x10000000
_V4L2_CAP_DEVICE_CAPS = 0x80000000


def _fourcc(a, b, c, d): return (ord(a) | (ord(b) << 8) | (ord(c) << 16) | (ord(d) << 24))


def _fourcc_be(a, b, c, d): return (_fourcc(a, b, c, d) | (1 << 31))


class V4L2_PIX_FMT(enum.IntEnum):
    RGB332 = _fourcc('R', 'G', 'B', '1')
    RGB444 = _fourcc('R', '4', '4', '4')
    ARGB444 = _fourcc('A', 'R', '1', '2')
    XRGB444 = _fourcc('X', 'R', '1', '2')
    RGB555 = _fourcc('R', 'G', 'B', 'O')
    ARGB555 = _fourcc('A', 'R', '1', '5')
    XRGB555 = _fourcc('X', 'R', '1', '5')
    RGB565 = _fourcc('R', 'G', 'B', 'P')
    RGB555X = _fourcc('R', 'G', 'B', 'Q')
    ARGB555X = _fourcc_be('A', 'R', '1', '5')
    XRGB555X = _fourcc_be('X', 'R', '1', '5')
    RGB565X = _fourcc('R', 'G', 'B', 'R')
    BGR666 = _fourcc('B', 'G', 'R', 'H')
    BGR24 = _fourcc('B', 'G', 'R', '3')
    RGB24 = _fourcc('R', 'G', 'B', '3')
    BGR32 = _fourcc('B', 'G', 'R', '4')
    ABGR32 = _fourcc('A', 'R', '2', '4')
    XBGR32 = _fourcc('X', 'R', '2', '4')
    RGB32 = _fourcc('R', 'G', 'B', '4')
    ARGB32 = _fourcc('B', 'A', '2', '4')
    XRGB32 = _fourcc('B', 'X', '2', '4')
    GREY = _fourcc('G', 'R', 'E', 'Y')
    Y4 = _fourcc('Y', '0', '4', ' ')
    Y6 = _fourcc('Y', '0', '6', ' ')
    Y10 = _fourcc('Y', '1', '0', ' ')
    Y12 = _fourcc('Y', '1', '2', ' ')
    Y16 = _fourcc('Y', '1', '6', ' ')
    Y16_BE = _fourcc_be('Y', '1', '6', ' ')
    Y10BPACK = _fourcc('Y', '1', '0', 'B')
    PAL8 = _fourcc('P', 'A', 'L', '8')
    UV8 = _fourcc('U', 'V', '8', ' ')
    YUYV = _fourcc('Y', 'U', 'Y', 'V')
    YYUV = _fourcc('Y', 'Y', 'U', 'V')
    YVYU = _fourcc('Y', 'V', 'Y', 'U')
    UYVY = _fourcc('U', 'Y', 'V', 'Y')
    VYUY = _fourcc('V', 'Y', 'U', 'Y')
    Y41P = _fourcc('Y', '4', '1', 'P')
    YUV444 = _fourcc('Y', '4', '4', '4')
    YUV555 = _fourcc('Y', 'U', 'V', 'O')
    YUV565 = _fourcc('Y', 'U', 'V', 'P')
    YUV32 = _fourcc('Y', 'U', 'V', '4')
    HI240 = _fourcc('H', 'I', '2', '4')
    HM12 = _fourcc('H', 'M', '1', '2')
    M420 = _fourcc('M', '4', '2', '0')
    NV12 = _fourcc('N', 'V', '1', '2')
    NV21 = _fourcc('N', 'V', '2', '1')
    NV16 = _fourcc('N', 'V', '1', '6')
    NV61 = _fourcc('N', 'V', '6', '1')
    NV24 = _fourcc('N', 'V', '2', '4')
    NV42 = _fourcc('N', 'V', '4', '2')
    NV12M = _fourcc('N', 'M', '1', '2')
    NV21M = _fourcc('N', 'M', '2', '1')
    NV16M = _fourcc('N', 'M', '1', '6')
    NV61M = _fourcc('N', 'M', '6', '1')
    NV12MT = _fourcc('T', 'M', '1', '2')
    NV12MT_16X16 = _fourcc('V', 'M', '1', '2')
    YUV410 = _fourcc('Y', 'U', 'V', '9')
    YVU410 = _fourcc('Y', 'V', 'U', '9')
    YUV411P = _fourcc('4', '1', '1', 'P')
    YUV420 = _fourcc('Y', 'U', '1', '2')
    YVU420 = _fourcc('Y', 'V', '1', '2')
    YUV422P = _fourcc('4', '2', '2', 'P')
    YUV420M = _fourcc('Y', 'M', '1', '2')
    YVU420M = _fourcc('Y', 'M', '2', '1')
    YUV422M = _fourcc('Y', 'M', '1', '6')
    YVU422M = _fourcc('Y', 'M', '6', '1')
    YUV444M = _fourcc('Y', 'M', '2', '4')
    YVU444M = _fourcc('Y', 'M', '4', '2')
    SBGGR8 = _fourcc('B', 'A', '8', '1')
    SGBRG8 = _fourcc('G', 'B', 'R', 'G')
    SGRBG8 = _fourcc('G', 'R', 'B', 'G')
    SRGGB8 = _fourcc('R', 'G', 'G', 'B')
    SBGGR10 = _fourcc('B', 'G', '1', '0')
    SGBRG10 = _fourcc('G', 'B', '1', '0')
    SGRBG10 = _fourcc('B', 'A', '1', '0')
    SRGGB10 = _fourcc('R', 'G', '1', '0')
    SBGGR10P = _fourcc('p', 'B', 'A', 'A')
    SGBRG10P = _fourcc('p', 'G', 'A', 'A')
    SGRBG10P = _fourcc('p', 'g', 'A', 'A')
    SRGGB10P = _fourcc('p', 'R', 'A', 'A')
    SBGGR10ALAW8 = _fourcc('a', 'B', 'A', '8')
    SGBRG10ALAW8 = _fourcc('a', 'G', 'A', '8')
    SGRBG10ALAW8 = _fourcc('a', 'g', 'A', '8')
    SRGGB10ALAW8 = _fourcc('a', 'R', 'A', '8')
    SBGGR10DPCM8 = _fourcc('b', 'B', 'A', '8')
    SGBRG10DPCM8 = _fourcc('b', 'G', 'A', '8')
    SGRBG10DPCM8 = _fourcc('B', 'D', '1', '0')
    SRGGB10DPCM8 = _fourcc('b', 'R', 'A', '8')
    SBGGR12 = _fourcc('B', 'G', '1', '2')
    SGBRG12 = _fourcc('G', 'B', '1', '2')
    SGRBG12 = _fourcc('B', 'A', '1', '2')
    SRGGB12 = _fourcc('R', 'G', '1', '2')
    SBGGR16 = _fourcc('B', 'Y', 'R', '2')
    MJPEG = _fourcc('M', 'J', 'P', 'G')
    JPEG = _fourcc('J', 'P', 'E', 'G')
    DV = _fourcc('d', 'v', 's', 'd')
    MPEG = _fourcc('M', 'P', 'E', 'G')
    H264 = _fourcc('H', '2', '6', '4')
    H264_NO_SC = _fourcc('A', 'V', 'C', '1')
    H264_MVC = _fourcc('M', '2', '6', '4')
    H263 = _fourcc('H', '2', '6', '3')
    MPEG1 = _fourcc('M', 'P', 'G', '1')
    MPEG2 = _fourcc('M', 'P', 'G', '2')
    MPEG4 = _fourcc('M', 'P', 'G', '4')
    XVID = _fourcc('X', 'V', 'I', 'D')
    VC1_ANNEX_G = _fourcc('V', 'C', '1', 'G')
    VC1_ANNEX_L = _fourcc('V', 'C', '1', 'L')
    VP8 = _fourcc('V', 'P', '8', '0')
    CPIA1 = _fourcc('C', 'P', 'I', 'A')
    WNVA = _fourcc('W', 'N', 'V', 'A')
    SN9C10X = _fourcc('S', '9', '1', '0')
    SN9C20X_I420 = _fourcc('S', '9', '2', '0')
    PWC1 = _fourcc('P', 'W', 'C', '1')
    PWC2 = _fourcc('P', 'W', 'C', '2')
    ET61X251 = _fourcc('E', '6', '2', '5')
    SPCA501 = _fourcc('S', '5', '0', '1')
    SPCA505 = _fourcc('S', '5', '0', '5')
    SPCA508 = _fourcc('S', '5', '0', '8')
    SPCA561 = _fourcc('S', '5', '6', '1')
    PAC207 = _fourcc('P', '2', '0', '7')
    MR97310A = _fourcc('M', '3', '1', '0')
    JL2005BCD = _fourcc('J', 'L', '2', '0')
    SN9C2028 = _fourcc('S', 'O', 'N', 'X')
    SQ905C = _fourcc('9', '0', '5', 'C')
    PJPG = _fourcc('P', 'J', 'P', 'G')
    OV511 = _fourcc('O', '5', '1', '1')
    OV518 = _fourcc('O', '5', '1', '8')
    STV0680 = _fourcc('S', '6', '8', '0')
    TM6000 = _fourcc('T', 'M', '6', '0')
    CIT_YYVYUY = _fourcc('C', 'I', 'T', 'V')
    KONICA420 = _fourcc('K', 'O', 'N', 'I')
    JPGL = _fourcc('J', 'P', 'G', 'L')
    SE401 = _fourcc('S', '4', '0', '1')
    S5C_UYVY_JPG = _fourcc('S', '5', 'C', 'I')
    Y8I = _fourcc('Y', '8', 'I', ' ')
    Y12I = _fourcc('Y', '1', '2', 'I')
    Z16 = _fourcc('Z', '1', '6', ' ')


class V4L2_BUF_TYPE(enum.IntEnum):
    VIDEO_CAPTURE = 1
    VIDEO_OUTPUT = 2
    VIDEO_OVERLAY = 3
    VBI_CAPTURE = 4
    VBI_OUTPUT = 5
    SLICED_VBI_CAPTURE = 6
    SLICED_VBI_OUTPUT = 7
    VIDEO_OUTPUT_OVERLAY = 8
    VIDEO_CAPTURE_MPLANE = 9
    VIDEO_OUTPUT_MPLANE = 10
    SDR_CAPTURE = 11
    SDR_OUTPUT = 12
    PRIVATE = 0x80


class V4L2_FRMSIZE_TYPE(enum.IntEnum):
    DISCRETE = 1
    CONTINUOUS = 2
    STEPWISE = 3


class V4L2_FRMIVAL_TYPE(enum.IntEnum):
    DISCRETE = 1
    CONTINUOUS = 2
    STEPWISE = 3


class V4L2_FIELD(enum.IntEnum):
    ANY = 0
    NONE = 1
    TOP = 2
    BOTTOM = 3
    INTERLACED = 4
    SEQ_TB = 5
    SEQ_BT = 6
    ALTERNATE = 7
    INTERLACED_TB = 8
    INTERLACED_BT = 9


V4L2_MODE_HIGHQUALITY = 0x0001
V4L2_CAP_TIMEPERFRAME = 0x1000


class V4L2_MEMORY(enum.IntEnum):
    MMAP = 1
    USERPTR = 2
    OVERLAY = 3
    DMABUF = 4


VideoPort = enum.Enum('VideoPort', 'CSI USB')


class VideoConfig(object):

    def __init__(self, width=640, height=480, pixel_format=V4L2_PIX_FMT.RGB24, framerate=30):
        self.pixel_format = pixel_format
        self.width = width
        self.height = height
        self.interval = fract()
        self.interval.numerator = 1
        self.interval.denominator = framerate


class Video(object):

    def __init__(self, device='/dev/video0', blocking=False):
        self.device = device
        flags = os.O_RDWR
        if not blocking:
            flags |= os.O_NONBLOCK
        self.device_fd = os.open(self.device, flags)
        self.converter = _v4lconvert.create(self.device_fd)

    def close(self):
        os.close(self.device_fd)

    def __enter__(self):
        return self

    def __exit__(self, ex_type, ex_value, trace):
        self.close()

    def _ioctl(self, request, arg):
        while True:
            result = _v4l2.ioctl(self.device_fd, request, arg)
            e = get_errno()
            if not (((-1 == result) and ((e == errno.EINTR) or (e == errno.EAGAIN)))):
                break
        return result

    def query_capability(self):
        cap = capability()
        result = self._ioctl(_VIDIOC.QUERYCAP, byref(cap))
        if not (cap.capabilities & _V4L2_CAP_VIDEO_CAPTURE):
            raise RuntimeError("The device doesn't support the single-planar API through the Video Capture interface.")
        if not (cap.capabilities & _V4L2_CAP_STREAMING):
            raise RuntimeError("The device doesn't support the streaming I/O method.")
        driver = ''.join(map(chr, itertools.takewhile(lambda x: x > 0, cap.driver)))
        if driver == 'bm2835 mmal':
            return VideoPort.CSI
        elif driver[:len('uvcvideo')] == 'uvcvideo':
            return VideoPort.USB
        else:
            raise RuntimeError("unknown driver '{}'".format(driver))

    def lookup_config(self, width, height, framerate, pixel_format, expected_format):

        results = []

        candidate = VideoConfig()

        def infinite_range():
            i = 0
            while True:
                yield i
                i += 1

        for i in infinite_range():

            fmt = fmtdesc()
            fmt.index = i
            fmt.type = V4L2_BUF_TYPE.VIDEO_CAPTURE

            result = _v4l2.ioctl(self.device_fd, _VIDIOC.ENUM_FMT, byref(fmt))
            if result != 0 and get_errno() == errno.EINVAL:
                break
            if result != 0:
                raise RuntimeError(errno.errorcode[get_errno()])

            if not fmt.pixelformat == pixel_format:
                continue

            candidate.pixel_format = fmt.pixelformat

            for j in infinite_range():
                frmsize = frmsizeenum()
                frmsize.index = j
                frmsize.pixel_format = fmt.pixelformat

                result = _v4l2.ioctl(self.device_fd, _VIDIOC.ENUM_FRAMESIZES, byref(frmsize))
                if result != 0 and get_errno() == errno.EINVAL:
                    break
                if result != 0:
                    raise RuntimeError(errno.errorcode[get_errno()])

                if frmsize.type == V4L2_FRMSIZE_TYPE.DISCRETE:
                    if width <= frmsize.discrete.width and height <= frmsize.discrete.height:
                        candidate.width = frmsize.discrete.width
                        candidate.height = frmsize.discrete.height
                    else:
                        continue
                elif frmsize.type == V4L2_FRMSIZE_TYPE.CONTINUOUS:
                    if frmsize.stepwise.min_width <= width and width <= frmsize.stepwise.max_width and \
                       frmsize.stepwise.min_height <= height and height <= frmsize.stepwise.max_height:
                        candidate.width = width
                        candidate.height = height
                    else:
                        continue
                elif frmsize.type == V4L2_FRMSIZE_TYPE.STEPWISE:
                    if frmsize.stepwise.min_width <= width and width <= frmsize.stepwise.max_width and \
                       frmsize.stepwise.min_height <= height and height <= frmsize.stepwise.max_height:
                        candidate.width = (width - frmsize.stepwise.min_width + frmsize.stepwise.step_width -
                                           1) // frmsize.stepwise.step_width * frmsize.stepwise.step_width + frmsize.stepwise.min_width
                        candidate.height = (height - frmsize.stepwise.min_height + frmsize.stepwise.step_height -
                                            1) // frmsize.stepwise.step_height * frmsize.stepwise.step_height + frmsize.stepwise.min_height
                    else:
                        continue
                else:
                    continue

                for k in infinite_range():
                    frmival = frmivalenum()
                    frmival.index = k
                    frmival.pixel_format = candidate.pixel_format
                    frmival.width = candidate.width
                    frmival.height = candidate.height

                    result = _v4l2.ioctl(self.device_fd, _VIDIOC.ENUM_FRAMEINTERVALS, byref(frmival))
                    if result != 0 and get_errno() == errno.EINVAL:
                        break
                    if result != 0:
                        raise RuntimeError(errno.errorcode[get_errno()])

                    if frmival.type == V4L2_FRMIVAL_TYPE.DISCRETE:
                        rate = frmival.discrete.denominator * 1. / frmival.discrete.numerator
                        if framerate <= rate:
                            candidate.interval = frmival.discrete
                        else:
                            continue
                    elif frmival.type == V4L2_FRMIVAL_TYPE.CONTINUOUS:
                        min_rate = frmival.stepwise.max.denominator * 1. / frmival.stepwise.max.numerator
                        max_rate = frmival.stepwise.min.denominator * 1. / frmival.stepwise.min.numerator
                        if min_rate <= framerate and framerate <= max_rate:
                            candidate.interval.numerator = 1
                            candidate.interval.denominator = framerate
                        else:
                            continue
                    elif frmival.type == V4L2_FRMIVAL_TYPE.STEPWISE:
                        min_rate = frmival.stepwise.max.denominator * 1. / frmival.stepwise.max.numerator
                        max_rate = frmival.stepwise.min.denominator * 1. / frmival.stepwise.min.numerator
                        step_rate = frmival.stepwise.step.denominator * 1. / frmival.stepwise.step.numerator
                        if min_rate <= framerate and framerate <= max_rate:
                            s = 0
                            while True:
                                mn = frmival.stepwise.max.numerator
                                md = frmival.stepwise.max.denominator
                                sn = frmival.stepwise.step.numerator
                                sd = frmival.stepwise.step.denominator
                                n = mn * sd - s * sn * md
                                d = md * sd
                                rate = d * 1.0 / n
                                if framerate <= rate:
                                    candidate.interval.numerator = n
                                    candidate.interval.denominator = d
                                    break
                        else:
                            continue
                    else:
                        continue

                    if expected_format == pixel_format:
                        results.append(candidate)
                    else:
                        if self.try_convert(candidate, candidate.width, candidate.height, expected_format) is not None:
                            results.append(candidate)

                    old_candidate = candidate
                    candidate = VideoConfig()
                    candidate.pixel_format = old_candidate.pixel_format
                    candidate.width = old_candidate.width
                    candidate.height = old_candidate.height

        return results

    def try_convert(self, conf, expected_width, expected_height, expected_format):

        fmt = format()
        fmt.type = V4L2_BUF_TYPE.VIDEO_CAPTURE
        fmt.fmt.pix.width = conf.width
        fmt.fmt.pix.height = conf.height
        fmt.fmt.pix.pixelformat = conf.pixel_format
        fmt.fmt.pix.field = V4L2_FIELD.INTERLACED

        expected_fmt = format()
        expected_fmt.type = V4L2_BUF_TYPE.VIDEO_CAPTURE
        expected_fmt.fmt.pix.width = expected_width
        expected_fmt.fmt.pix.height = expected_height
        expected_fmt.fmt.pix.pixelformat = expected_format
        expected_fmt.fmt.pix.field = V4L2_FIELD.INTERLACED

        result = _v4lconvert.try_format(self.converter, byref(expected_fmt), byref(fmt))
        if -1 == result:
            raise RuntimeError("incompatible format")

        before = (expected_width, expected_height, expected_format)
        after = (expected_fmt.fmt.pix.width,
                 expected_fmt.fmt.pix.height,
                 expected_fmt.fmt.pix.pixelformat)

        if before == after:
            return (fmt, expected_fmt)
        else:
            return None

    def set_format(self, conf, expected_width=None, expected_height=None, expected_format=None):

        if expected_width is None:
            expected_width = conf.width
        if expected_height is None:
            expected_height = conf.height
        if expected_format is None:
            expected_format = conf.pixel_format

        fmts = self.try_convert(conf, expected_width, expected_height, expected_format)
        if fmts is None:
            fmts = self.try_convert(conf, conf.width, conf.height, expected_format)
            if fmts is None:
                raise RuntimeError("incompatible format")

        (self.fmt, self.expected_fmt) = fmts
        result = self._ioctl(_VIDIOC.S_FMT, byref(self.fmt))
        if -1 == result:
            raise RuntimeError("ioctl(VIDIOC_S_FMT)")

        return (self.expected_fmt.fmt.pix.width,
                self.expected_fmt.fmt.pix.height,
                self.expected_fmt.fmt.pix.pixelformat)

    def set_framerate(self, conf):

        parm = streamparm()
        parm.type = V4L2_BUF_TYPE.VIDEO_CAPTURE
        parm.parm.capture.timeperframe = conf.interval
        parm.parm.capture.capturemode = V4L2_MODE_HIGHQUALITY

        result = self._ioctl(_VIDIOC.S_PARM, byref(parm))
        if -1 == result:
            raise RuntimeError("ioctl(VIDIOC_S_PARM)")

        return True

    def set_rotation(self, rotation):
        """
        Rotate video if it supported.

        Args:
            rotation (int): video rotation (must be multiples of 90)

        Returns:
            boolean: result
        """

        assert rotation % 90 == 0, 'rotation must be multiples of 90'
        rot = (rotation // 90) % 4 * 90  # [0, 90, 180, 270]

        qctrl = queryctrl()
        qctrl.id = V4L2_CID.ROTATE
        result = self._ioctl(_VIDIOC.QUERYCTRL, byref(qctrl))
        if -1 == result:
            return False

        ctrl = control()
        ctrl.id = V4L2_CID.ROTATE
        ctrl.value = rot

        result = self._ioctl(_VIDIOC.S_CTRL, byref(ctrl))
        if -1 == result:
            return False

        return True

    def set_horizontal_flip(self, flip):
        """
        Flip video (horizontal) if it supported.

        Args:
            flip (boolean): enable flip

        Returns:
            boolean: result
        """

        qctrl = queryctrl()
        qctrl.id = V4L2_CID.HFLIP
        result = self._ioctl(_VIDIOC.QUERYCTRL, byref(qctrl))
        if -1 == result:
            return False
        if qctrl.flags & V4L2_CTRL_FLAG_DISABLED:
            return False

        ctrl = control()
        ctrl.id = V4L2_CID.HFLIP
        ctrl.value = flip

        result = self._ioctl(_VIDIOC.S_CTRL, byref(ctrl))
        if -1 == result:
            return False
        expected = ctrl.value
        result = self._ioctl(_VIDIOC.G_CTRL, byref(ctrl))
        if -1 == result:
            return False
        if expected != ctrl.value:
            return False

        return True

    def set_vertical_flip(self, flip):
        """
        Flip video (vertical) if it supported.

        Args:
            flip (boolean): enable flip

        Returns:
            boolean: result
        """

        qctrl = queryctrl()
        qctrl.id = V4L2_CID.VFLIP
        result = self._ioctl(_VIDIOC.QUERYCTRL, byref(qctrl))
        if -1 == result:
            return False
        if qctrl.flags & V4L2_CTRL_FLAG_DISABLED:
            return False

        ctrl = control()
        ctrl.id = V4L2_CID.VFLIP
        ctrl.value = flip

        result = self._ioctl(_VIDIOC.S_CTRL, byref(ctrl))
        if -1 == result:
            return False
        expected = ctrl.value
        result = self._ioctl(_VIDIOC.G_CTRL, byref(ctrl))
        if -1 == result:
            return False
        if expected != ctrl.value:
            return False

        return True

    def set_exposure_time(self, ms=None):
        """
        Set exposure time.

        Args:
            ms (int or None): exposure time [msec] (None means auto)

        Returns:
            boolean: result
        """

        if ms is None:
            ctrl = control()
            ctrl.id = V4L2_CID.EXPOSURE_AUTO
            ctrl.value = V4L2_EXPOSURE.AUTO
            result = self._ioctl(_VIDIOC.S_CTRL, byref(ctrl))
            if -1 == result:
                return False
            expected = ctrl.value
            result = self._ioctl(_VIDIOC.G_CTRL, byref(ctrl))
            if -1 == result:
                return False
            if expected != ctrl.value:
                return False
        else:
            ctrl = control()
            ctrl.id = V4L2_CID.EXPOSURE_AUTO
            ctrl.value = V4L2_EXPOSURE.MANUAL
            result = self._ioctl(_VIDIOC.S_CTRL, byref(ctrl))
            if -1 == result:
                return False
            expected = ctrl.value
            result = self._ioctl(_VIDIOC.G_CTRL, byref(ctrl))
            if -1 == result:
                return False
            if expected != ctrl.value:
                return False

            ctrl.id = V4L2_CID.EXPOSURE_ABSOLUTE
            ctrl.value = int(10 * ms)  # [100us]
            result = self._ioctl(_VIDIOC.S_CTRL, byref(ctrl))
            if -1 == result:
                return False
            expected = ctrl.value
            result = self._ioctl(_VIDIOC.G_CTRL, byref(ctrl))
            if -1 == result:
                return False
            if expected != ctrl.value:
                return False

        return True

    def request_buffers(self, n):

        req = requestbuffers()
        req.count = n
        req.type = V4L2_BUF_TYPE.VIDEO_CAPTURE
        req.memory = V4L2_MEMORY.MMAP

        result = self._ioctl(_VIDIOC.REQBUFS, byref(req))
        if -1 == result:
            raise RuntimeError("ioctl(VIDIOC_REQBUFS): {}".format(errno.errorcode[get_errno()]))

        return [VideoBuffer._from_query(self, i) for i in range(n)]

    def queue_buffer(self, video_buf):

        result = self._ioctl(_VIDIOC.QBUF, byref(video_buf.buf))
        if -1 == result:
            raise RuntimeError("ioctl(VIDIOC_QBUF): {}".format(errno.errorcode[get_errno()]))

        return True

    def start_streaming(self):

        cap = c_int(V4L2_BUF_TYPE.VIDEO_CAPTURE)
        result = self._ioctl(_VIDIOC.STREAMON, byref(cap))
        if -1 == result:
            raise RuntimeError("ioctl(VIDIOC_STREAMON): {}".format(errno.errorcode[get_errno()]))

        return VideoStream(self)

    def stop_streaming(self):

        cap = c_int(V4L2_BUF_TYPE.VIDEO_CAPTURE)
        result = self._ioctl(_VIDIOC.STREAMOFF, byref(cap))
        if -1 == result:
            raise RuntimeError("ioctl(VIDIOC_STREAMON): {}".format(errno.errorcode[get_errno()]))

        return True

    def dequeue_buffer(self, timeout=1):

        class FDWrapper:

            def __init__(self, fd):
                self.fd = fd

            def fileno(self):
                return self.fd

        rlist, _, _ = select.select([FDWrapper(self.device_fd)], [], [], timeout)
        if len(rlist) == 0:
            raise RuntimeError("Capture timeout")

        buf = buffer()
        buf.type = V4L2_BUF_TYPE.VIDEO_CAPTURE
        buf.memory = V4L2_MEMORY.MMAP
        result = self._ioctl(_VIDIOC.DQBUF, byref(buf))
        if -1 == result:
            raise RuntimeError("ioctl(VIDIOC_DQBUF): {}".format(errno.errorcode[get_errno()]))

        return VideoBuffer(self, buf)

    def requeue_buffer(self, video_buf):

        result = self._ioctl(_VIDIOC.QBUF, byref(video_buf.buf))
        if -1 == result:
            raise RuntimeError("ioctl(VIDIOC_QBUF): {}".format(errno.errorcode[get_errno()]))


class VideoStream(object):

    def __init__(self, video):
        self.video = video

    def __enter__(self):
        return self

    def __exit__(self, ex_type, ex_value, trace):
        self.video.stop_streaming()

    def capture(self, timeout=1, in_expected_format=True):

        buf = self.video.dequeue_buffer(timeout=timeout)
        mapped_buf = buf.get_mapped_buffer()

        dst = bytes(self.video.expected_fmt.fmt.pix.sizeimage)
        if in_expected_format:
            _v4lconvert.convert(self.video.converter,
                                byref(self.video.fmt),
                                byref(self.video.expected_fmt),
                                mapped_buf,
                                self.video.fmt.fmt.pix.sizeimage,
                                cast(dst, POINTER(c_uint8)),
                                self.video.expected_fmt.fmt.pix.sizeimage)
        else:
            stream = io.BytesIO(dst)
            stream.write(mapped_buf.contents)

        buf.unmap_buffer()
        self.video.requeue_buffer(buf)

        return dst


class VideoBuffer(object):

    @classmethod
    def _from_query(cls, video, index):

        buf = buffer()
        buf.type = V4L2_BUF_TYPE.VIDEO_CAPTURE
        buf.memory = V4L2_MEMORY.MMAP
        buf.index = index

        set_errno(0)
        result = video._ioctl(_VIDIOC.QUERYBUF, byref(buf))
        if -1 == result:
            raise RuntimeError("ioctl(VIDIOC_QYERYBUF): {}".format(errno.errorcode[get_errno()]))

        return cls(video, buf)

    def __init__(self, video, buf):
        self.video = video
        self.buf = buf
        self.mapped_buf = None

    def get_mapped_buffer(self):
        if self.mapped_buf is not None:
            return self.mapped_buf
        result = _v4l2.mmap(None, self.buf.length,
                            mmap.PROT_READ | mmap.PROT_WRITE, mmap.MAP_SHARED,
                            self.video.device_fd, c_longlong(self.buf.m.offset))
        if result == -1:
            raise RuntimeError("mmap failed: {}".format(errno.errorcode[get_errno()]))
        self.mapped_buf = cast(result, POINTER(ARRAY(c_uint8, self.buf.length)))
        return self.mapped_buf

    def unmap_buffer(self):
        if self.mapped_buf is None:
            return
        result = _v4l2.munmap(self.mapped_buf, self.buf.length)
        if result == -1:
            raise RuntimeError("munmap failed: {}".format(errno.errorcode[get_errno()]))
        self.mapped_buf = None

# if __name__ == '__main__':

#     from PIL import Image
#     import traceback

#     with Video('/dev/video1') as video:
#         print("video.query_capability():",video.query_capability())
#         candidates = video.lookup_config(640, 480, 30, V4L2_PIX_FMT.RGB24) + video.lookup_config(640, 480, 30, V4L2_PIX_FMT.YUYV)
#         for i, config in enumerate(candidates):
#             print("video.lookup_config index:", i,
#                   V4L2_PIX_FMT(config.pixel_format), config.width, config.height,
#                   config.interval.denominator / config.interval.numerator)
#         config = candidates[0]
#         print("video.set_format(config):",video.set_format(config, 320, 240, V4L2_PIX_FMT.RGB24))
#         print("video.set_framerate(config):",video.set_framerate(config))
#         print("video.set_rotation(90):", video.set_rotation(90))
#         num_of_buffers = 2
#         buffers = video.request_buffers(num_of_buffers)
#         print("video.request_buffers({}):".format(num_of_buffers), buffers)
#         for buf in buffers:
#             print("video.queue_buffer({}):".format(buf), video.queue_buffer(buf))
#         with video.start_streaming() as stream:
#             try:
#                 for i in range(5):
#                     dat = stream.capture()
#                     img = Image.fromarray(np.frombuffer(dat, dtype=np.uint8).reshape(240,320,3))
#                     img.save('{:03d}.png'.format(i))
#             except:
#                 print(traceback.format_exc())



import io
import os
import socket
import base64
from threading import Lock
import traceback
from .task import Isolated


def _read_tokens(conn, n):
    result = []
    s = b''
    x = n
    while x > 0:
        c = conn.recv(1)
        if c == b' ':
            x -= 1
            result.append(s)
            s = b''
        else:
            s += c
    return result


def _read_bytes(conn, n):
    result = b''
    while len(result) < n:
        result += conn.recv(1024)
    return result


class CommandServer(Isolated):

    """Actcast Command Server

    This server handles these commands

    * 'Take Photo'
        * responses cached image as png data

    """

    def __init__(self, sock_path=None):
        super(CommandServer, self).__init__()
        self.sock_path = None
        env = 'ACTCAST_COMMAND_SOCK'
        if env in os.environ:
            self.sock_path = os.environ[env]
        if sock_path is not None:
            self.sock_path = sock_path
        self.running = True
        self.img_lock = Lock()
        self.img = None

    def run(self):
        """Run and start the activity"""
        if self.sock_path is None:
            return
        s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        try:
            os.unlink(self.sock_path)
        except FileNotFoundError as e:
            pass  # ignore
        s.bind(self.sock_path)
        s.settimeout(1)
        s.listen(1)
        while self.running:
            # Wait photo
            while self.running:
                with self.img_lock:
                    if self.img is None:
                        continue
                    else:
                        break
            try:
                conn, addr = s.accept()
                [request_id, command_id, command_data_length] = map(int, _read_tokens(conn, 3))
                command_data = _read_bytes(conn, command_data_length)
                if command_id == 0:  # Take Photo
                    header = "data:image/png;base64,"
                    with self.img_lock:
                        pngimg = io.BytesIO()
                        self.img.save(pngimg, format='PNG')
                        b64img = base64.b64encode(pngimg.getbuffer())
                    conn.sendall("{} {} {} {}{}\n".format(request_id, 0, len(header) +
                                                          len(b64img), header, b64img.decode('utf-8')).encode('utf-8'))
                else:
                    conn.sendall("{} 2 0\n".format(request_id))
                conn.close()
            except socket.timeout:
                pass
            except Exception as e:
                traceback.print_exc()
        os.remove(self.sock_path)

    def update_image(self, image):
        """

        Update the cached 'Take Photo' command image.

        Args:
            image (:class:`~PIL.Image`): image

        """
        with self.img_lock:
            self.img = image.copy()

    def stop(self):
        """Stop the activity"""
        self.running = False



from ctypes import *


class capability(Structure):
    _fields_ = [
        ('driver', c_ubyte * 16),
        ('card', c_ubyte * 32),
        ('bus_info', c_ubyte * 32),
        ('version', c_uint),
        ('capabilities', c_uint),
        ('device_caps', c_uint),
        ('reserved', c_uint * 3),
    ]


class fmtdesc(Structure):
    _fields_ = [
        ('index', c_uint),
        ('type', c_uint),
        ('flags', c_uint),
        ('description', c_ubyte * 32),
        ('pixelformat', c_uint),
        ('reserved', c_uint * 4),
    ]


class frmsize_discrete(Structure):
    _fields_ = [
        ('width', c_uint),
        ('height', c_uint),
    ]


class frmsize_stepwise(Structure):
    _fields_ = [
        ('min_width', c_uint),
        ('max_width', c_uint),
        ('step_width', c_uint),
        ('min_height', c_uint),
        ('max_height', c_uint),
        ('step_height', c_uint),
    ]


class _frmsize_for_frmsizeenum(Union):
    _fields_ = [
        ('discrete', frmsize_discrete),
        ('stepwise', frmsize_stepwise),
    ]


class frmsizeenum(Structure):
    _anonymous_ = (
        '_frmsize',
    )
    _fields_ = [
        ('index', c_uint),
        ('pixel_format', c_uint),
        ('type', c_uint),
        ('_frmsize', _frmsize_for_frmsizeenum),
        ('reserved', c_uint * 2),
    ]


class fract(Structure):
    _fields_ = [
        ('numerator', c_uint),
        ('denominator', c_uint),
    ]


class frmival_stepwise(Structure):
    _fields_ = [
        ('min', fract),
        ('max', fract),
        ('step', fract),
    ]


class _frmival_for_frmivalenum(Union):
    _fields_ = [
        ('discrete', fract),
        ('stepwise', frmival_stepwise),
    ]


class frmivalenum(Structure):
    _anonymous_ = (
        '_frmival',
    )
    _fields_ = [
        ('index', c_uint),
        ('pixel_format', c_uint),
        ('width', c_uint),
        ('height', c_uint),
        ('type', c_uint),
        ('_frmival', _frmival_for_frmivalenum),
        ('reserved', c_uint * 2),
    ]


class pix_format(Structure):
    _fields_ = [
        ('width', c_uint),
        ('height', c_uint),
        ('pixelformat', c_uint),
        ('field', c_uint),
        ('bytesperline', c_uint),
        ('sizeimage', c_uint),
        ('colorspace', c_uint),
        ('priv', c_uint),
        ('flags', c_uint),
        ('ycbcr_enc', c_uint),
        ('quantization', c_uint),
        ('xfer_func', c_uint),
    ]


class plane_pix_format(Structure):
    _fields_ = [
        ('sizeimage', c_uint),
        ('bytesperline', c_uint),
        ('reserved', c_ushort * 6),
    ]


class pix_format_mplane(Structure):
    _fields_ = [
        ('width', c_uint),
        ('height', c_uint),
        ('pixelformat', c_uint),
        ('field', c_uint),
        ('colorspace', c_uint),
        ('plane_fmt', plane_pix_format * 8),  # VIDEO_MAX_PLANES
        ('num_planes', c_ubyte),
        ('flags', c_ubyte),
        ('ycbcr_enc', c_ubyte),
        ('quantization', c_ubyte),
        ('xfer_func', c_ubyte),
        ('reserved', c_ubyte * 7),
    ]


class rect(Structure):
    _fields_ = [
        ('left', c_int),
        ('top', c_int),
        ('width', c_uint),
        ('height', c_uint),
    ]


class clip(Structure):
    pass


clip._fields_ = [
    ('c', rect),
    ('next', POINTER(clip)),
]


class window(Structure):
    _fields_ = [
        ('w', rect),
        ('field', c_uint),
        ('chromakey', c_uint),
        ('clips', POINTER(clip)),
        ('clipcount', c_uint),
        ('bitmap', c_void_p),
        ('global_alpha', c_ubyte),
    ]


class vbi_format(Structure):
    _fields_ = [
        ('sampling_rate', c_uint),
        ('offset', c_uint),
        ('samples_per_line', c_uint),
        ('sample_format', c_uint),
        ('start', c_int * 2),
        ('count', c_uint * 2),
        ('flags', c_uint),
        ('reserved', c_uint * 2),
    ]


class sliced_vbi_format(Structure):
    _fields_ = [
        ('service_lines', 2 * (24 * c_ushort)),
        ('io_size', c_uint),
        ('reserved', c_uint * 2),
    ]


class sdr_format(Structure):
    _fields_ = [
        ('pixelformat', c_uint),
        ('buffersize', c_uint),
        ('reserved', c_ubyte * 24),
    ]


class _fmt_for_format(Union):
    _fields_ = [
        ('pix', pix_format),
        ('pix_mp', pix_format_mplane),
        ('win', window),
        ('vbi', vbi_format),
        ('sliced', sliced_vbi_format),
        ('sdr', sdr_format),
        ('raw_data', c_ubyte * 200),
    ]


class format(Structure):
    _fields_ = [
        ('type', c_uint),
        ('fmt', _fmt_for_format),
    ]


class captureparm(Structure):
    _fields_ = [
        ('capability', c_uint),
        ('capturemode', c_uint),
        ('timeperframe', fract),
        ('extendedmode', c_uint),
        ('readbufferbs', c_uint),
        ('reserved', c_uint * 4),
    ]


class outputparm(Structure):
    _fields_ = [
        ('capability', c_uint),
        ('outputmode', c_uint),
        ('timeperframe', fract),
        ('extendedmode', c_uint),
        ('writebuffers', c_uint),
        ('reserved', c_uint * 4),
    ]


class _parm_for_streamparm(Union):
    _fields_ = [
        ('capture', captureparm),
        ('output', outputparm),
        ('raw_data', c_ubyte * 200),
    ]


class streamparm(Structure):
    _fields_ = [
        ('type', c_uint),
        ('parm', _parm_for_streamparm),
    ]


class control(Structure):
    _fields_ = [
        ('id', c_uint),
        ('value', c_int),
    ]


class queryctrl(Structure):
    _fields_ = [
        ('id', c_uint),
        ('type', c_uint),
        ('name', c_ubyte * 32),
        ('minimum', c_int),
        ('maximum', c_int),
        ('step', c_int),
        ('default_value', c_int),
        ('flags', c_uint),
        ('reserved', c_uint * 2),
    ]


class requestbuffers(Structure):
    _fields_ = [
        ('count', c_uint),
        ('type', c_uint),
        ('memory', c_uint),
        ('reserved', c_uint * 2),
    ]


class timeval(Structure):
    _fields_ = [
        ('sec', c_long),
        ('usec', c_long),
    ]


class timecode(Structure):
    _fields_ = [
        ('type', c_uint),
        ('flags', c_uint),
        ('frames', c_ubyte),
        ('seconds', c_ubyte),
        ('minutes', c_ubyte),
        ('hours', c_ubyte),
        ('userbits', c_ubyte * 4),
    ]


class _m_for_plane(Union):
    _fields_ = [
        ('mem_offset', c_uint),
        ('userptr', c_ulong),
        ('fd', c_int),
    ]


class plane(Structure):
    _fields_ = [
        ('bytesused', c_uint),
        ('length', c_uint),
        ('m', _m_for_plane),
        ('data_offset', c_uint),
        ('reserved', c_uint * 11),
    ]


class _m_for_buffer(Union):
    _fields_ = [
        ('offset', c_uint),
        ('userptr', c_ulong),
        ('planes', POINTER(plane)),
        ('fd', c_int),
    ]


class buffer(Structure):
    _fields_ = [
        ('index', c_uint),
        ('type', c_uint),
        ('bytesused', c_uint),
        ('flags', c_uint),
        ('field', c_uint),

        ('timestamp', timeval),
        ('timecode', timecode),
        ('sequence', c_uint),

        ('memory', c_uint),
        ('m', _m_for_buffer),
        ('length', c_uint),
        ('reserved2', c_uint),
        ('reserved', c_uint),
    ]



from .pipe import Pipe


class Producer(Pipe):

    """Producer Task."""

    def __init__(self):
        """
        """
        super(Producer, self).__init__()

    def _add_in_queue(self, q):
        raise NotImplementedError('This is producer')

    def _inlet(self):
        while self._is_running():
            yield ()



from .pipe import Pipe


class Consumer(Pipe):

    """Consumer Task."""

    def __init__(self):
        super(Consumer, self).__init__()

    def _add_out_queue(self, q):
        raise NotImplementedError('This is consumer')

    def _outlet(self, o):
        raise NotImplementedError('This is consumer')

    def run(self):
        """Run and start the activity"""
        for i in self._inlet():
            self.proc(i)
            if not self._is_running:
                break

    def connect(self, follow):
        """
        Raises error

        Args:
            follow (:class:`~actfw_core.task.Task`): following task
        """
        if not issubclass(type(follow), Task):
            raise TypeError("type(follow) must be a subclass of actfw_core.task.Task.")
        raise NotImplementedError('This is consumer')



import os
import signal
import time
import json
from actfw_core.task import Task


class Application:

    """Actcast Application"""

    def __init__(self):
        self.running = True
        signal.signal(signal.SIGINT, self._handler)
        signal.signal(signal.SIGTERM, self._handler)
        self.tasks = []
        self.settings = None
        env = 'ACT_SETTINGS_PATH'
        if env in os.environ:
            try:
                with open(os.environ[env]) as f:
                    self.settings = json.load(f)
            except FileNotFoundError:
                pass

    def _handler(self, sig, frame):
        self.running = False

    def get_settings(self, default_settings):
        """

        Get given Act settings.

        Args:
            default_settings (dict): default settings

        Returns:
            dict: updated settings

        Notes:
            Copy default_settings and overwrite it by given Act settings.

        """
        if not isinstance(default_settings, dict):
            raise TypeError("default_settings must be dict.")
        settings = default_settings.copy()
        if self.settings is not None:
            settings.update(self.settings)
        return settings

    def register_task(self, task):
        """

        Register the application task.

        Args:
            task (:class:`~actfw_core.task.Task`): task
        """
        if not issubclass(type(task), Task):
            raise TypeError("type(task) must be a subclass of actfw_core.task.Task.")
        self.tasks.append(task)

    def run(self):
        """Start application"""
        for task in self.tasks:
            task.start()

        try:
            while self.running:
                time.sleep(1)
        except KeyboardInterrupt:
            pass
        except:
            raise

        for task in self.tasks:
            task.stop()
        for task in self.tasks:
            task.join()



import os
import json
from pathlib import Path
import actfw_core.task
import actfw_core.capture
from .application import Application
from .command_server import CommandServer

from actfw_core import _version

__version__ = _version.__version__


def notify(notification, *args, **kwargs):
    """

    Make a notification to Actcast.

    Args:
        notification (list of dict): dicts must be encodable to JSON.

    Example:

        >>> import actfw_core
        >>> actfw_core.notify([{'msg': 'Hello!'}])
        [{"msg": "Hello!"}]

    """
    if type(notification) != list:
        raise TypeError('must be a list of JSON encodable objects.')
    kwargs['flush'] = True
    print(json.dumps(notification), *args, **kwargs)


_default_heartbeat_file = Path('/root/heartbeat')


def _default_heartbeat(*args, **kwargs):
    _default_heartbeat_file.touch()


_heartbeat_function = _default_heartbeat


def set_heartbeat_function(f):
    """

    Set heartbeat action.

    Args:
        f (function): function which execute heartbeat action

    Example:

        >>> import actfw_core
        >>> def heartbeat(): print("working!")
        ...
        >>> actfw_core.set_heartbeat_function(heartbeat)
        >>> actfw_core.heartbeat()
        working!

    """
    global _heartbeat_function
    _heartbeat_function = f


def heartbeat(*args, **kwargs):
    """

    Execute heartbeat action.

    Notes:
        Default action is 'touch /root/heartbeat'.

    """
    _heartbeat_function(*args, **kwargs)



from threading import Thread
from queue import Queue, Empty, Full
import traceback
import inspect
from .task import Task


class Pipe(Task):

    """Straightforward Pipeline Task."""

    def __init__(self):
        """
        """
        super(Pipe, self).__init__()
        self.running = True
        self.in_queues = []
        self.out_queues = []
        self.out_queue_id = 0

    def _is_running(self):
        return self.running

    def _add_in_queue(self, q):
        self.in_queues.append(q)

    def _add_out_queue(self, q):
        self.out_queues.append(q)

    def _inlet(self):
        in_queue_id = 0
        length = len(self.in_queues)
        while self._is_running():
            try:
                i = self.in_queues[in_queue_id].get(timeout=1)
                yield i
                in_queue_id = (in_queue_id + 1) % length
            except Empty:
                pass
            except GeneratorExit:
                break
            except:
                traceback.print_exc()

    def _outlet(self, o):
        length = len(self.out_queues)
        while self._is_running():
            try:
                self.out_queues[self.out_queue_id].put(o, timeout=1)
                self.out_queue_id = (self.out_queue_id + 1) % length
                return True
            except Full:
                pass
            except:
                traceback.print_exc()
        return False

    def run(self):
        """Run and start the activity"""
        for i in self._inlet():
            o = self.proc(i)
            self._outlet(o)
            if not self._is_running():
                break

    def proc(self, i):
        """
        Pipeline Task Processor

        Args:
            i : task input
        """
        raise NotImplementedError("'proc' must be overridden.")

    def stop(self):
        """Stop the activity"""
        self.running = False

    def connect(self, follow):
        """
        Connect following task.

        Args:
            follow (:class:`~actfw_core.task.Task`): following task
        """
        q = Queue(1)
        follow._add_in_queue(q)
        self._add_out_queue(q)
